/*->c.xstr  */


#include <stdafx.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>
#include <locale.h>



#include "os.h"
#include "flex.h"

#include "deb.h"

#include "xext.h"
#include "xint.h"





/*****************************************************************************/
/* x string manager */


typedef struct string
{
 int offset;
 int maxlen;
 int curlen;
} string;


static char   * stringspace;
static int      stringsize;

static string * stringmap;
int      nostrings;



char * stringptr(int s)
{
 int offset;
 offset=stringmap[s].offset;
 return(stringspace+offset);
}




static void stringmem(void)
{
 if(zlevel) zraise(ZMEM,"");
 else       xraise(ERRMEM);
}


int mkstring(int len)
{
 if(!flex_chunk((flex_ptr)&stringmap,(nostrings+1)*sizeof(string),STRMAPCHUNK))
 {
  stringmap[nostrings++].offset=stringsize;

  stringmap[nostrings-1].curlen=len;
  len++;
  len=(len+0x3)&(~0x3);
  stringmap[nostrings-1].maxlen=len;

  stringsize+=len;

  if(flex_chunk((flex_ptr)&stringspace,stringsize,STRCHUNK)) stringmem();
 }
 else stringmem();

 return(nostrings-1);
}



static void chklen(int s,int newlen)
{
 int len;
 int offset;
 int delta;

 offset=stringmap[s].offset;
 len=stringmap[s].maxlen;

 if(newlen==len) return;

 delta=newlen-len;

 stringmap[s].maxlen+=delta;

 if(delta>0)
 {
  if(flex_chunk((flex_ptr)&stringspace,stringsize+delta,STRCHUNK)) 
                                                                stringmem();
 }

 memmove(stringspace+offset+len+delta,stringspace+offset+len,
                                      stringsize-offset-len);

 while(++s<nostrings)
 {
  if(stringmap[s].offset>=0) stringmap[s].offset+=delta;
 }

 if(delta<0) flex_chunk((flex_ptr)&stringspace,stringsize+delta,STRCHUNK);
 stringsize+=delta;

}



static void chktruelen(int s,int len)
{
 len++;
 len=(len+0x3)&(~0x3);
 chklen(s,len);
}




void rmstring(int s)
{
 int len;
 int offset;

 if(s<0 || s>=nostrings) 
 {
/* dprintf(0,"string error s=%d nos=%d\n",s,nostrings); */
  return;
 }

 offset=stringmap[s].offset;
 len=stringmap[s].maxlen;

 if(s==(nostrings-1))
 {
  while(1)
  {
   nostrings--;
   s--;

/*   printf("rem loop s=%d nostrings=%d offset=%d",s,nostrings,stringmap[s].offset); */

   if(s<0) break;
   if(stringmap[s].offset>=0) break;
  }
 }
 else
 if(s<(nostrings-1))
 {
  /* shift memory from the end of this string to end of table down by len */
  memmove(stringspace+offset,stringspace+offset+len,stringsize-offset-len);
  stringmap[s].offset=-1;
  while(++s<nostrings)
  {
   if(stringmap[s].offset>=0) stringmap[s].offset-=len;
  }
 }

 stringsize-=len;

/* printf("rmstring stringsize=%d - %d\n",stringsize,len); */

 flex_chunk((flex_ptr)&stringspace,stringsize,STRCHUNK);
 flex_chunk((flex_ptr)&stringmap,nostrings*sizeof(string),STRMAPCHUNK);
}




/* trashes top n strings on string stack, optionally saving top n */

void stringpull(int n,int top)
{
 int i;

 i=nostrings-1-top;

 while(n--)
 {
  while(i>=0)
  {
   if(stringmap[i].offset>=0) break; 
   else i--;
  }

  rmstring(i--);
 }
}



/*
void stringpull(int n,int top)
{
 int i;

 i=nostrings-1-top;

 while(1)
  if(stringmap[i].offset>=0) break; 
  else i--;

 while(n--) rmstring(i--);
}
*/





void setstrings(int n)
{

 while(nostrings>n) rmstring(nostrings-1);
}



/* create a new string */

int createstring(char * p)
{
 int s;

 s=mkstring(strlen(p));
 strcpy(stringptr(s),p);
 return(s);
}


int createstringr(char ** p)
{
 int s;

 s=mkstring(strlen(*p));
 strcpy(stringptr(s),*p);

 return(s);
}


// size excludes terminating zero
// it is assumed there is a terminating zero
// at the memory pointed to by *p

int createstringrlen(char ** p,int size)
{
 int s;

 s=mkstring(size);
 memcpy(stringptr(s),*p,size+1);

 return(s);
}




/* creates a string and then copies a string into it */

int duplicatestring(int s1)
{
 char * p;
 char * q;
 int    s2;
 int    len;

 len=stringmap[s1].curlen;

 s2=mkstring(len);
 p=stringptr(s1);
 q=stringptr(s2);
 memcpy(q,p,len+1);
 stringmap[s2].curlen=len;

 return(s2);
}




/* assigns one string to another */

void copystring(int s1,int s2)
{
 char * p;
 char * q;

/* printf("copy string oldlen=%d newlen=%d\n",stringmap[s1].len,stringmap[s2].len); */

 chklen(s1,stringmap[s2].maxlen);
 p=stringptr(s1);
 q=stringptr(s2);

/* printf("after copy string oldlen=%d newlen=%d\n",stringmap[s1].len,stringmap[s2].len); */

 memcpy(p,q,stringmap[s2].maxlen);
 stringmap[s1].curlen=stringmap[s2].curlen;
}






os_error * setstring(int s,char * p)
{
 int    len;

 len=strlen(p);
 chktruelen(s,len);
 stringmap[s].curlen=len;

 return(NULL);
}



os_error * assignstring(int s,char * p)
{
 setstring(s,p);
 strcpy(stringptr(s),p);
 return(NULL);
}





os_error * appendstring(int s,char * p)
{
 int len1;
 int len2;
 char * q;

 len1=strlen(p);

 q=stringptr(s);
 len2=stringmap[s].curlen;

 chktruelen(s,len1+len2);

 q=stringptr(s);
 strcpy(q+len2,p);

 stringmap[s].curlen=len1+len2;

 return(NULL);
}







/* the string s2 is appended to s1 */

void catstrings(int s1,int s2)
{
 char * p;
 char * q;
 int    len1;
 int    len2;

 p=stringptr(s1);
 q=stringptr(s2);

 len1=stringmap[s1].curlen;
 len2=stringmap[s2].curlen;

/* printf("len1=%d len2=%d\n",len1,len2);
 printf("maplen1=%d maplen2=%d\n",stringmap[s1].len,stringmap[s2].len); */


 chktruelen(s1,len1+len2);


/* printf("after maplen1=%d maplen2=%d\n",stringmap[s1].len,stringmap[s2].len);
*/

 p=stringptr(s1);
 q=stringptr(s2);

 memcpy(p+len1,q,len2);
 *(p+len1+len2)=0;

 stringmap[s1].curlen=len1+len2;
}



/* the string s1 is multiplied n times */

void mulstring(int s1,int n)
{
 char * p;
 int    len;
 int    i;

 p=stringptr(s1);

 len=strlen(p);
 chktruelen(s1,len*n);
 p=stringptr(s1);

 for(i=1;i<n;i++) memcpy(p+i*len,p,len);
 *(p+n*len)=0;
 stringmap[s1].curlen=len*n;
}




int divstring(int s1,int s2)
{
 char * p;
 char * q;

 p=stringptr(s1);
 q=stringptr(s2);

 q=strstr(p,q);
 if(q) return(q-p);
 else  return(-1);
}



int modstring(int s1,int s2)
{
 char * p;
 char * q;

 p=stringptr(s1);
 q=stringptr(s2);

 return(strspn(p,q));
}



int cmpstrings(int s1,int s2)
{
 char * p;
 char * q;
 p=stringptr(s1);
 q=stringptr(s2);

 return(strcmp(p,q));
}



/* the string s1 is made upper case */

void upperstring(int s1)
{
 char * p;
 int    len;
 int    i;
 p=stringptr(s1);
 len=strlen(p);
 for(i=0;i<len;i++) p[i]=xtoup(p[i]);
}



/* the string s1 is made lower case */

void lowerstring(int s1)
{
 char * p;
 int    len;
 int    i;
 p=stringptr(s1);
 len=strlen(p);
 for(i=0;i<len;i++) p[i]=xtolo(p[i]);
}




/* the case of string s1 is swopped */

void swopstring(int s1)
{
 char * p;
 int    len;
 int    i;
 p=stringptr(s1);
 len=strlen(p);
 for(i=0;i<len;i++) 
  if(xislo(p[i])) p[i]=xtoup(p[i]);
  else
  if(xisup(p[i])) p[i]=xtolo(p[i]);
}




/* string s1 is shifted left */

void shiftstringleft(int s1,int n)
{
 char * p;
 int    len;

 p=stringptr(s1);
 len=stringmap[s1].curlen;

 if(n<0) n=0;
 else
 if(n>len) n=len;

 memmove(p,p+n,len-n+1);
 chktruelen(s1,len-n);
 stringmap[s1].curlen=len-n;
}


/* string s1 is shifted right */

void shiftstringright(int s1,int n)
{
 char * p;
 int    len;

 p=stringptr(s1);
 len=stringmap[s1].curlen;
 if(n<0) n=0;
 else
 if(n>len) n=len;

 p[len-n]=0;
 chktruelen(s1,len-n);
 stringmap[s1].curlen=len-n;
}


/* string s1 has byte catted at right */

/*
void catstringintright(int s1,int byte)
{
 char * p;
 int    len;

 p=stringptr(s1);
 len=strlen(p);
 chktruelen(s1,len+1);
 p[len]=byte;
 p[len+1]=0;
}
 */

/* string s1 has byte catted at left */

/*
void catstringintleft(int s1,int byte)
{
 char * p;
 int    len;

 p=stringptr(s1);
 len=strlen(p);
 chktruelen(s1,len+1);
 memmove(p,p+1,len+1);
 p[0]=byte;
}
 */


/* length of string */

int stringlen(int s1)
{
/* return(strlen(stringptr(s1))); */
 return(stringmap[s1].curlen);
}



/* nth. char in string */

int stringn(int s1,int n)
{
 char * p;
 int    len;

 p=stringptr(s1);
 len=stringmap[s1].curlen;

 if(n<0) n=0;
 if(n>len) n=len;

 return(*(p+n));
}



/* called once boots string table */

os_error * xstringstart(void)
{
 os_error * err;
 err=flex_alloc((flex_ptr)&stringspace,STRCHUNK);
 if(!err) err=flex_alloc((flex_ptr)&stringmap,STRMAPCHUNK);
 stringsize=0;
 nostrings=0;
 return(err);
}




xcasefn xtoup=toupper;
xcasefn xtolo=tolower;
xcasefn xisup=isupper;
xcasefn xislo=islower;




void xsetcasefns(xcasefn toup,xcasefn tolo,xcasefn isup,xcasefn islo)
{
 xtoup=toup;
 xtolo=tolo;
 xisup=isup;
 xislo=islo;
}











#ifdef NEVER







void stringdump(int fp)
{
 int i;
 char temp[128];
 static int count;

 if((clock()>1000) && ((count++ & 0xFFFF)==0))
 {
  dprintf(0,"nostrings=%d stringsize=%d\n",nostrings,stringsize); 

  for(i=0;i<nostrings;i++)
  {
   xstrncpy(temp,stringspace+stringmap[i].offset,120);

   dprintf(0,"s=%d offset=%d curlen=%d maxlen=%d %s\n",
      i,stringmap[i].offset,stringmap[i].curlen,stringmap[i].maxlen,temp);
  }
 }
}







void stringdump(int fp)
{
 int i;

 dprintf(0,"nostrings=%d stringsize=%d\n",nostrings,stringsize); 

 for(i=0;i<nostrings;i++)
 {
  dprintf(0,"s=%d offset=%d curlen=%d maxlen=%d str=%s\n",
     i,stringmap[i].offset,stringmap[i].curlen,stringmap[i].maxlen,
                            stringspace+stringmap[i].offset);
 }
}





void stringdump0(char * s)
{
 dprintf(0,"%s str=%s\n",s,stringspace+stringmap[0].offset);
}



#endif


