/*->c.chsect */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <ctype.h>


#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"

#include "err.h"
#include "transform.h"
#include "etc.h"
#include "poll.h"


#include "constants.h"
#include "str.h"
#include "object.h"
#include "chain.h"
#include "undo.h"
#include "frameflow.h"

#include "reslink.h"

#include "view.h"
#include "viewo.h"
#include "file.h"
#include "select.h"

#include "txstr.h"
#include "fl.h"
#include "text.h"

#include "page.h"
#include "new.h"
#include "newx.h"
#include "xpage.h"





#include "chsect.h"



#define UCHSECTF  0x7300
#define UCHSECTS  0x7301
#define UCHADD    0x7302
#define UCHREM    0x7303
#define UCHFRAME  0x7304
#define UCHLINK   0x7305
#define UCHUNLINK 0x7306




os_error * recalcmaxpagesize(filestr * file,int branch)
{
 sectionstr * sectionp;
 int          section;
 int          maxw;
 int          maxh;


 if(branch<0)
 {
  recalcmaxpagesize(file,MST);
  recalcmaxpagesize(file,DOC);
 }
 else
 {
  maxw=maxh=0;

  section=file->hdr.section[branch];
  while(section)
  {
   sectionp=(sectionstr*)fob_find(section,file);
   section=sectionp->chain.next;

   maxw=MAX(maxw,sectionp->data.width);
   maxh=MAX(maxh,sectionp->data.height);
  }

  file->hdr.maxpagewidth[branch]=maxw;
  file->hdr.maxpageheight[branch]=maxh;
 }

 return(NULL);
}




os_error * changepagesize(filestr * file,int section)
{
 sectionstr      * sectionp;
 int               page;
 sectiondatastr    msect;
 int               message;

 sectionp=(sectionstr*)fob_find(section,file);
 page=sectionp->chain.child;
 msect=sectionp->data;

 message=OM_PAGESIZE;
 objectchainmessage(page,file,&message,&msect);

 return(NULL);
}





static os_error * checksection(filestr * file,int section)
{
 os_error      * err;
 sectionstr    * sectionp;
 int             hand;
 sectiondatastr  sect;
 int             page;
 pagestr       * pagep;

 sectionp=(sectionstr*)fob_find(section,file);
 sect=sectionp->data;
 page=sectionp->chain.child;

 for(hand=0;hand<2;hand++)
 {
  if(hand<=sect.dble)
  {
   if(sect.mainkey[hand] && 
      !objectinobject(sect.mainkey[hand],page,file)) 
                                                sect.mainkey[hand]=0;

   if(hand==0)
   {
    pagep=(pagestr*)fob_find(page,file);
    page=pagep->chain.next;
   }
  }
  else sect.mainkey[hand]=0;
 }

 err=fob_mod(section,file,(fobjptr)&sectionp);
 sectionp->data=sect;

 return(err);
}






static sectiondatastr osect;
static sectiondatastr msect;
static int            msection;







/* tagged to UCHADD  */
/* remove the page   */

static os_error * unchadd(filestr * file)
{
 int          page;
 pagestr    * pagep;
 sectionstr * sectionp;

 undoread(file,&page,sizeof(int));
 undowritefn(file,UCHREM,&page,sizeof(int));

 file->hdr.np[MST]--;
 sectionp=(sectionstr*)fob_find(msection,file);
 sectionp->np--;

 pagep=(pagestr*)fob_find(page,file);
 page=pagep->chain.prev;

 fob_mod(page,file,(fobjptr)&pagep);
 pagep->chain.next=0;

 return(NULL);
}



/* tagged to UCHREM  */
/* reinsert the page */

static os_error * unchremscrap(filestr * file)
{
 int page;
 undoread(file,&page,sizeof(int));
 return(scrappagecontents(page,file));
}


static os_error * unchrem(filestr * file)
{
 int          page;
 int          prev;
 pagestr    * pagep;
 sectionstr * sectionp;

 undoread(file,&page,sizeof(int));
 undowritefn(file,UCHADD,&page,sizeof(int));

 file->hdr.np[MST]++;
 sectionp=(sectionstr*)fob_find(msection,file);
 sectionp->np++;

 pagep=(pagestr*)fob_find(page,file);
 prev=pagep->chain.prev;

 fob_mod(prev,file,(fobjptr)&pagep);
 pagep->chain.next=page;

 return(NULL);
}



/* tagged to UCHFRAME */

static os_error * unchframe(filestr * file)
{
 int text;
 undoread(file,&text,sizeof(int));
 return(objectdelete((objhdrstr*)fob_find(text,file),0,file));
}







#ifdef NEVER

 if(msect.mainkey[0] && msect.mainkey[1])  /* link frames */
 {
  textlink(msect.mainkey[0],msect.mainkey[1],file);
 }
 else
 if(msect.mainkey[0])
 {
  fob_mod(msect.mainkey[0],file,(fobjptr)&textp);
  textp->fchain.prev=textp->fchain.next=0;
 }
 else
 if(msect.mainkey[1])
 {
  fob_mod(msect.mainkey[1],file,(fobjptr)&textp);
  textp->fchain.next=textp->fchain.prev=0;
 }

#endif






/* called last */

/* tagged with UCHSECTS */

static os_error * chsects(filestr * file)
{
 os_error   * err;
 sectionstr * sectionp;
 int          message;
 linkstr      link;
 pagestr    * pagep;


 err=NULL;


 if(!osect.dble && msect.dble && !undoforward(file) && undoactive(file))
 {
  /* relink */
  message=OM_MREMAKELINK;
  sectionp=(sectionstr*)fob_find(msection,file);
  link.mpage=sectionp->chain.child;
  pagep=(pagestr*)fob_find(link.mpage,file);
  link.mpage=pagep->chain.next;
  link.msection=msection;
  pagep=(pagestr*)fob_find(link.mpage,file);
  err=objectchainmessage(pagep->chain.child,file,&message,&link);
 }

 if(osect.width!=msect.width || osect.height!=msect.height)
 {
  changepagesize(file,msection);
 }


 frflend(file);

 fileresetviews(file,FRVREFRESH|FRVMST);

 undowritefn(file,UCHSECTF,&msection,sizeof(int));
 undowrite(file,&osect,sizeof(sectiondatastr));

 return(err);
}




static os_error * chsectfsub(filestr * file)
{
 os_error   * err;
 sectionstr * sectionp;
 int          message;
 linkstr      link;
 pagestr    * pagep;

 checksection(file,msection);

 err=undowritefn(file,UCHSECTS,NULL,0);

 fob_mod(msection,file,(fobjptr)&sectionp);
 osect=sectionp->data;
 sectionp->data=msect;

 if(osect.dble && !msect.dble && undoforward(file))
 {
  /* break links */
  message=OM_MBREAKLINK;
  link.mpage=sectionp->chain.child;
  link.msection=msection;
  pagep=(pagestr*)fob_find(link.mpage,file);
  err=objectchainmessage(pagep->chain.child,file,&message,&link);
 }


 if(osect.width!=msect.width || osect.height!=msect.height)
 {
  recalcmaxpagesize(file,MST);
 }

 frflstart(file);

 return(err);
}



/* called first */

/* tagged with UCHSECTF */

static os_error * chsectf(filestr * file)
{
 undoread(file,&msection,sizeof(int));
 undoread(file,&msect,sizeof(sectiondatastr));

 return(chsectfsub(file));
}





static os_error * chlink(filestr * file)
{
 textlink(msect.mainkey[0],msect.mainkey[1],file);
 return(undowritefn(file,UCHUNLINK,NULL,0));
}



static os_error * chunlink(filestr * file)
{
 textboxstr * textp;
 fob_mod(osect.mainkey[0],file,(fobjptr)&textp);
 textp->fchain.next=0;
 return(undowritefn(file,UCHLINK,NULL,0));
}




static os_error * chsectframes(filestr * file,int section,newstr * xnew)
{
 os_error   * err;
 int          hand;
 int          page[2];
 pagestr    * pagep;
 sectionstr * sectionp;       
 boxstr       box;
 objectloc    loc;
 newpagestr   newpage;
 textboxstr * textp;


 err=NULL;


 sectionp=(sectionstr*)fob_find(section,file);
 page[0]=sectionp->chain.child;
 if(osect.dble)
 {
  pagep=(pagestr*)fob_find(page[0],file);
  page[1]=pagep->chain.next;
 }

 memset(&loc,0,sizeof(objectloc));
 loc.file=file;
 loc.branch=1;
 loc.chain=0;
 loc.after=0;
 loc.section=section;


 if(osect.dble && msect.dble)
 {
  if(osect.mainkey[0] && osect.mainkey[1] && 
      (!xnew->principleframe[0] || !xnew->principleframe[1]))
  {
   fob_mod(osect.mainkey[0],file,(fobjptr)&textp);
   if(textp->fchain.next==osect.mainkey[1])
   {
    textp->fchain.next=0;
    undowritefn(file,UCHLINK,NULL,0);
   }
  }
 }


 /* update as many principle frames as you need to */

 for(hand=0;hand<=osect.dble && hand<=msect.dble;hand++)
 {
  if(xnew->principleframe[hand] && osect.mainkey[hand])
  {
   box.x=xnew->leftmargin[hand];
   box.w=xnew->sectiondata.width-xnew->leftmargin[hand]-xnew->rightmargin[hand];
   box.y=xnew->topmargin[hand];
   box.h=xnew->sectiondata.height-xnew->topmargin[hand]-xnew->botmargin[hand];

   textchangebox(osect.mainkey[hand],&box,&xnew->columndata[hand],
                                           xnew->columns[hand],
                                           xnew->headermargin[hand],
                                           xnew->footermargin[hand],file);

   msect.mainkey[hand]=osect.mainkey[hand];
  }
  else
  if(xnew->principleframe[hand] && !osect.mainkey[hand])
  {
   /* create a main frame */

   loc.parent=loc.page=page[hand];
   loc.copymode=LOCNEW;     /* new source created */

   err=createpageboxes(&loc,xnew,hand);

   msect.mainkey[hand]=xnew->sectiondata.mainkey[hand];
   undowritefn(file,UCHFRAME,&msect.mainkey[hand],sizeof(int));
  }
  else
  if(!xnew->principleframe[hand] && osect.mainkey[hand])
  {
   /* remove a main frame */

   /* if(isselected(osect.mainkey[hand],file)) clearselectionlo(); */

   err=objectdelete((objhdrstr*)fob_find(osect.mainkey[hand],file),0,file);
   msect.mainkey[hand]=0;
  }
  else /* if(!new->principleframe[hand] && !osect.mainkey[hand]) */
  {
   msect.mainkey[hand]=0;
  }
 }

 /* get right number of pages */

 if(osect.dble && !msect.dble)
 {
  /* remove second page */

 /* if(isselectedpage(page[1],file)) clearselectionlo(); */

  undowritefn(file,UCHREM,&page[1],sizeof(int));

  file->hdr.np[MST]--;
  sectionp=(sectionstr*)fob_find(loc.section,file);
  sectionp->np--;

  fob_mod(page[0],file,(fobjptr)&pagep);
  pagep->chain.next=0;

  msect.mainkey[1]=0;
 }
 else
 if(!osect.dble && msect.dble)
 {
  /* create second page */

  newpage.data=msect;
  newpage.pagenumber=1;  /* right hand page */

  err=objectnew(&loc,OPAGE,&newpage);
  if(!err)
  {
   loc.page=loc.object;

   file->hdr.np[MST]++;

   sectionp=(sectionstr*)fob_find(loc.section,file);
   sectionp->np++;

   pagep=(pagestr*)fob_find(loc.page,file);
   pagep->section=loc.section;

   pagep->chain.prev=page[0];

   pagep=(pagestr*)fob_find(page[0],file);
   pagep->chain.next=loc.page;

   loc.parent=loc.page;
   loc.copymode=LOCNEW;     /* new source created */

   if(xnew->principleframe[1])
   {
    err=createpageboxes(&loc,xnew,1);
    msect.mainkey[1]=xnew->sectiondata.mainkey[1];
   }

   undowritefn(file,UCHADD,&loc.page,sizeof(int));
  }
 }

 fob_mod(section,file,(fobjptr)&sectionp);
 sectionp->data=msect;


 if(msect.dble && msect.mainkey[0] && msect.mainkey[1] && 
     (!osect.mainkey[0] || !osect.dble || !osect.mainkey[1]))
 {
  textp=(textboxstr*)fob_find(msect.mainkey[0],file);
  if(textp->data.repeat)
  {
   textp=(textboxstr*)fob_find(msect.mainkey[1],file);
   if(textp->data.repeat)
   {
    textlink(msect.mainkey[0],msect.mainkey[1],file);
    undowritefn(file,UCHUNLINK,NULL,0);
   }
  }
 }

 return(err);
}






os_error * changesection(newstr * xnew,int section,filestr * file)
{
 os_error   * err;

 err=modfile(file,section);
 if(!err)
 {
  err=undostart(file);
  if(!err)
  {
   clearselectionff(file,section,-1,1);

   msection=section;
   msect=xnew->sectiondata;

   chsectfsub(file);

   chsectframes(file,section,xnew);

   chsects(file);

   undoend(file);
  }
 }

 return(err);
}




#define UNDOS 7


static undofnstr undos[UNDOS]=
{
 chsectf,        NULL,          UCHSECTF,        "{UNCS}",
 chsects,        NULL,          UCHSECTS,        "{UNCS}",
 unchadd,        NULL,          UCHADD,          NULL,
 unchrem,        unchremscrap,  UCHREM,          NULL,
 unchframe,      NULL,          UCHFRAME,        NULL,
 chlink,         NULL,          UCHLINK,         NULL,
 chunlink,       NULL,          UCHUNLINK,       NULL,
};



os_error * chsectinit(void)
{
 return(undoaddblock(undos,UNDOS));
}



