/*->c.rulers */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "bbc.h"
#include "font.h"
#include "akbd.h"
#include "transform.h"

#include "err.h"
#include "xext.h"
#include "poll.h"
#include "temp.h"
#include "alloc.h"
#include "fsx.h"
#include "dbhi.h"
#include "trans.h"
#include "xmath.h"
#include "colour.h"
#include "key.h"
#include "units.h"
#include "deb.h"
#include "osunits.h"

#include "etc.h"

#include "str.h"
#include "constants.h"

#include "mhi.h"
#include "reslink.h"
#include "object.h"

#include "file.h"
#include "page.h"

#include "viewo.h"
#include "viewr.h"
#include "view.h"
#include "posn.h"
#include "xredraw.h"
#include "tools.h"
#include "toolbox.h"
#include "select.h"
#include "undo.h"
#include "infopal.h"
#include "caret.h"

#include "gline.h"

#include "rulers.h"




#define HMARGIN   40
#define VMARGIN   40




/* size of steps on rulers in our units */

int hsteps[MAXZONE][3]=
                  { /* mm         in      pts                      */
                    5*720000/254,18000,  24000, /* zone 0 25 -49%  */
                    2*720000/254, 9000,  12000, /* zone 1 50 -100% */
                    1*720000/254, 4500,   6000, /* zone 2 100-149% */


                  };

/* no of steps for each time value written */
     
int hbsteps[MAXZONE][3]=
                  {
                    10,           8,       6,   /* zone 0 25 -49%  */
                    10,           8,       6,   /* zone 1 50 -100% */
                    10,          16,      12,   /* zone 2 100-149% */
                  };




char * hpattern[MAXZONE][3]=
                  {
                   "8222242222",  "82224222",         "822422",
                   "8222242222",  "82224222",         "822422",
                   "8222242222",  "8232323242323232", "823232423232",
                  };

/* size of increment in unit each time value written  */

int ubsteps[MAXZONE][3]=
                  {
                   50,            2,     144,   /* zone 0 25 -49%  */
                   20,            1,      72,   /* zone 1 50 -100% */
                   10,            1,      72,   /* zone 2 100-149% */
                  };



/* return the marking zone for current zone */

int mzone(viewstr * view)
{
 int p;
 int measure;

 measure=view->data.units;

 p=100*view->data.zoom.mul;
 p/=view->data.zoom.div;

 if(measure==MILIMS)
 {
  if(p<=59)  p=0;
  else
  if(p<=119) p=1;
  else       p=2;
 }
 else
 if(measure==INCHES)
 {
  if(p<=39) p=0;
  else
  if(p<=74) p=1;
  else      p=2;
 }
 else
 if(measure==POINTS)
 {
  if(p<=39) p=0;
  else
  if(p<=59) p=1;
  else      p=2;
 }
 else
 {
  /* unknown unit case - only fixed in 2.47 */
  p=0;
 }

 return(p);
}

static int rulerfonthandle;

os_error * rulerfontstart(void)
{
 os_error * err;

 err=font_find("Arial",0,8*16,0,0,&rulerfonthandle);
 if(!err) err=font_setcolour(rulerfonthandle,PALTRANSPARENT,PALBLACK,0);

 return(err);
}


void rulerfontend(void)
{
 if(rulerfonthandle)
 {
  font_lose(rulerfonthandle);
  rulerfonthandle=0;
 }
}


void rulerprint(int n,int x,int y,int justify)
{
 char      string[8];
 int       flags;


 if(n<0) n=-n;

 sprintf(string,"%d",n % 1000);

 if(justify & 8) flags=FA_TOP;
 else
 if(justify & 4) flags=FA_BOTTOM;
 else            flags=FA_BASELINE;

 if(justify & 1) flags|=FA_RIGHT;        /* rh justify */
 if(justify & 2) flags|=FA_CENTRE;       /* centre     */

 font_align(flags);

 font_paint(string,font_ABS|font_OSCOORDS,x,y);
}               


/****************************************************************************/

static os_error * redrawvrulersub(viewredrawstr * v,int more);
static os_error * redrawhrulersub(viewredrawstr * v,int more);



static filestr * rmfile;
static int       rmpage;
static int       rmupdate;
static rmstr     rmold;
static rmstr     rmcur;
static int       rmbranch;


static os_error * updaterulers(void)
{
 os_error      * err;
 viewstr       * view;
 viewredrawstr   v;
 windowstr       window;
 int             more;
 int             xup;
 int             yup;

 err=NULL;

 initviewredraw(&v);

 xup=(rmcur.ax[0] || rmcur.ax[1] || rmold.ax[0] || rmold.ax[1]);
 yup=(rmcur.ay[0] || rmcur.ay[1] || rmold.ay[0] || rmold.ay[1]);

 if(rmfile && (xup || yup))
 {
  rmupdate=1;

  view=rmfile->view;

  while(view && !err)
  {
   if(view->branch==rmbranch && view->data.rulers)
   {
    v.view=view;
    v.file=view->file;

    if(yup)
    {
     v.redrawstr.w=view->vruler;

     if(v.redrawstr.w)
     {
      getw(v.redrawstr.w,&window);

      v.redrawstr.box.x0=window.x0-window.bx;
      v.redrawstr.box.x1=window.x1-window.bx;
      v.redrawstr.box.y0=window.y0-window.by;
      v.redrawstr.box.y1=window.y1-window.by;

               err=wimp_update_wind(&v.redrawstr,&more);
      if(!err) err=redrawvrulersub(&v,more);
     }
    }

    if(xup)
    {
     v.redrawstr.w=view->hruler;

     if(v.redrawstr.w)
     {
      getw(v.redrawstr.w,&window);

      v.redrawstr.box.x0=window.x0-window.bx;
      v.redrawstr.box.x1=window.x1-window.bx;
      v.redrawstr.box.y0=window.y0-window.by;
      v.redrawstr.box.y1=window.y1-window.by;

               err=wimp_update_wind(&v.redrawstr,&more);
      if(!err) err=redrawhrulersub(&v,more);
     }
    }

   }
   view=view->next;
  }

  rmupdate=0;
 }
 return(err);
}



os_error * rulermarkupdate(rmstr * xnew)
{
 os_error * err;
 int        i;


 rmold=rmcur;
 rmcur=*xnew;

 for(i=0;i<2;i++)
 {
  if(rmcur.ay[i] && rmold.ay[i] && rmcur.y[i]==rmold.y[i])
                                                  rmcur.ay[i]=rmold.ay[i]=0;
 }

 for(i=0;i<2;i++)
 {
  if(rmcur.ax[i] && rmold.ax[i] && rmcur.x[i]==rmold.x[i])
                                                  rmcur.ax[i]=rmold.ax[i]=0;
 }

 err=updaterulers();

 rmcur=*xnew;

 return(err);
}

os_error * rulermarkstart(rmstr * xnew,int page,int branch,filestr * file)
{
 memset(&rmold,0,sizeof(rmstr));
 rmfile=file;
 rmpage=page;
 rmbranch=branch;
 rmcur=*xnew;
 return(updaterulers());
}


os_error * rulermarkend(void)
{
 os_error * err;

 err=NULL;


 memset(&rmold,0,sizeof(rmstr));
 err=updaterulers();

 rmfile=NULL;

 return(err);
}



/* the problem here is that we have markers on the rulers, but...     */
/* sometimes when the window is opened the ruler associated page will */
/* change, and we don't know when that is - solution always remove    */
/* marks, and it flickers                                             */


os_error * rulersave(filestr * file)
{
 if(file==rmfile)
 {
  rmold=rmcur;
  memset(&rmcur,0,sizeof(rmstr));
  updaterulers();
 }

 return(NULL);
}


os_error * rulerload(filestr * file)
{
 if(file==rmfile)
 {
  rmcur=rmold;
  memset(&rmold,0,sizeof(rmstr));
  updaterulers();
 }

 return(NULL);
}


/*****************************************************************************/
                  /***** redraw vertical ruler *****/


static os_error * redrawvgaps(viewredrawstr * v,viewstr * view,
                                                int page1,int page2)
{
 os_error * err;
 pagestr  * pagep;
 int        height1;
 int        height2;
 int        height;

 wimp_setcolour(WIMPGREY4);

 height1=height2=0;

 if(page1)
 {
  pagep=(pagestr*)fob_find(page1,v->file);
  height1=pagep->box.h;
 }

 if(page2)
 {
  pagep=(pagestr*)fob_find(page2,v->file);
  height2=pagep->box.h;
 }

 height=MAX(height1,height2);
 height=v->ypo+height+view->border.botpb;


 /* inter page gaps */

   err=bbc_rectanglefill(v->ox,
                         v->oy-osunits(v->yps,view),
                               VMARGIN,
                    -osunits((v->ypo-v->yps)-view->border.toppb,view));

   err=bbc_rectanglefill(v->ox,v->oy-osunits(height,view),
                         VMARGIN,-osunits(v->ype-height,view));

 return(err);
}


static os_error * vrulermark(int a,int y,viewredrawstr * v)
{
 if(a)
 {
  y=v->oy-osunits(y+v->ypo,v->view);
  bbc_move(v->ox,y);
  return(bbc_draw(v->ox+VMARGIN,y));
 }

 return(NULL);
}



static os_error * redrawvrulersub(viewredrawstr * v,int more)
{
 os_error       * err;
 int              u;
 int              x;
 int              y;
 int              n;
 int              mz;
 int              measure;

 int              hstep;
 int              hbstep;
 int              ubstep;
 char    *        pattern;

 viewstr *        view;
 filestr *        file;

 int              ghost;
 pagestr        * page;
 sectionstr     * section;
 int              origin;
 int              page1;
 int              page2;
 int              nomorepages;

 err=NULL;

 view=v->view;
 file=v->file;

 measure=view->data.units;

 hstep=hsteps[mz=mzone(view)][measure];
 hbstep=hbsteps[mz][measure];
 ubstep=ubsteps[mz][measure];
 pattern=hpattern[mz][measure];

 while(more)
 {
  nomorepages=0;

  v->ox=v->redrawstr.box.x0-v->redrawstr.scx;
  v->oy=v->redrawstr.box.y1-v->redrawstr.scy;
  
  if(!rmupdate)
  {
   // guaranteed ruler window is VMARGIN pixels wide
//   x=v->ox+xpix2os(xos2pix(VMARGIN));
			x=v->ox+xpix2os(xos2pix(v->redrawstr.box.x1-v->redrawstr.box.x0)-1);
   bbc_move(x,v->redrawstr.box.y0-cvdu.extray);
   bbc_draw(x,v->redrawstr.box.y1+cvdu.extray);
  }

  v->ylox=-(v->redrawstr.g.y1-v->oy);
  v->yhix=-(v->redrawstr.g.y0-v->oy);
  v->xlox=v->redrawstr.g.x0-v->ox;
  v->xhix=v->redrawstr.g.x1-v->ox;

  v->yloz=ourunits(v->ylox,view);
  v->yhiz=ourunits(v->yhix,view);
  v->xloz=ourunits(v->xlox,view);
  v->xhiz=ourunits(v->xhix,view);

  v->page=view->page;
  v->section=view->section;
  v->yps=view->ps;

  firstxcoord(v);
  setxcoord(v);
  setycoord(v);
                             
  do
  {
   if(!(v->xps>v->xhiz || v->xpe<v->xloz || v->yps>v->yhiz || v->ype<v->yloz))
   {  
    if(v->yhi<v->ype) v->yhi+=hstep*hbstep;

    origin=v->ypo+view->data.rulery;
    n=(v->ylo-origin)/(hstep*hbstep);
    if(v->ylo<origin) n--;
    v->ylo=origin+n*hstep*hbstep;

    u=0;
    n*=ubstep;

    if(!rmupdate)
    {
     wimp_setcolour(WIMPBLACK);

     for(y=v->ylo;y<=v->yhi+hstep;y+=hstep)
     {
      if((y>=v->yps && y<=v->ype))
      {
     /*  if(u==0) bbc_move(v.ox,v.oy-osunits(y,view));
       else     bbc_move(v.ox+3*(VMARGIN >> 2),v.oy-osunits(y,view));

       bbc_draw(v.ox+VMARGIN,v.oy-osunits(y,view)); 
     */

       bbc_move(v->ox+VMARGIN,v->oy-osunits(y,view));
       bbc_draw(v->ox+(VMARGIN-((pattern[u]-'0')*VMARGIN)/PATTERNSCALE),
                            v->oy-osunits(y,view));
       if(u==0)
       {
       /* centre number on this mark and above */
         // rulerprint(n,v->ox+(VMARGIN/16),ycoordpix2os(ycoordos2pix(v->oy-osunits(y,view))-1),4);
        if(n) rulerprint(n,v->ox+(VMARGIN/16),(v->oy-osunits(y,view)),4);   
        n+=ubstep;
       }
      }
      else
      {
       if(u==0) n+=ubstep;
      }

      if(++u==hbstep) u=0;
     }
    }

    if(rmfile==file && v->page==rmpage)
    {
     seteorcol(WIMPWHITE,WIMPRED);

     

     if(rmupdate)
     {
      vrulermark(rmold.ay[0],rmold.y[0],v);
      vrulermark(rmold.ay[1],rmold.y[1],v);
     }

     vrulermark(rmcur.ay[0],rmcur.y[0],v);
     vrulermark(rmcur.ay[1],rmcur.y[1],v);
    }
   }

   page1=page2=0;

   ghost=isghost(view,v->page,v->section,0);
   if(!ghost) page1=v->page;

   if(view->data.dble)
   {
    page=(pagestr*)fob_find(v->page,file);
    section=(sectionstr*)fob_find(v->section,file);

    if(!ghost)
    {
     if(page->chain.next || section->chain.next)
     {
      if(page->chain.next) v->page=page->chain.next;
      else
      {
       v->section=section->chain.next;
       section=(sectionstr*)fob_find(v->section,file);
       v->page=section->chain.child;
      }
     }
     else
     {
      nomorepages=1;
     }
    }
    ghost=isghost(view,v->page,v->section,1);
    if(!ghost) page2=v->page;
   }


   if(!rmupdate) redrawvgaps(v,view,page1,page2);

   if(v->yhi>=v->yhiz) break;  /* there isn't going to be another page */
   if(nomorepages) break;


   nextycoord(v);

   page=(pagestr*)fob_find(v->page,file);
   section=(sectionstr*)fob_find(v->section,file);

   if(!ghost)
   {
    if(page->chain.next || section->chain.next)
    {
     if(page->chain.next) v->page=page->chain.next;
     else
     {
      v->section=section->chain.next;
      section=(sectionstr*)fob_find(v->section,file);
      v->page=section->chain.child;
     }
    }
    else
    {
     break;
    }
   }

   setycoord(v);

  } while(v->ylo<v->yhiz);


  wimp_get_rectangle(&v->redrawstr,&more);
 }

 return(err);
}



static os_error * redrawvruler(int handle,int userhandle)
{
 os_error      * err;
 viewredrawstr   v;
 int             more;

 initviewredraw(&v);

 v.redrawstr.w=handle;
 v.view=(viewstr*)userhandle;
 v.file=v.view->file;

 err=wimp_redraw_wind(&v.redrawstr,&more);
 if(!err)
 {
		if(more)
		{
   rulerfontstart();
   err=redrawvrulersub(&v,more);
   rulerfontend();
		}
 }

 return(err);
}



/*****************************************************************************/
                  /***** redraw horizontal ruler *****/


static os_error * redrawhgaps(viewredrawstr * v,viewstr * view)
{
 os_error * err;
 int        leftpb;
 int        rightpb;
 int        hand;
/* pagestr  * pagep; */
 int        width;

 err=NULL;

 if(view->data.dble)
 {
  wimp_setcolour(WIMPGREY4);

  /* left hand */

  hand=(v->xps>0);

  if(view->data.pasteboard)
  {
   if(hand && view->data.dble) leftpb=view->border.mid2pb;
   else                        leftpb=view->border.leftpb;
  }
  else                         leftpb=0;

  err=bbc_rectanglefill(v->ox+osunits(v->xps,view),
                        v->oy,
                          osunits(v->xpo-v->xps-leftpb,view),
                        -HMARGIN);

  /* right hand */

  if(view->data.pasteboard)
  {
   if(!hand && view->data.dble) rightpb=view->border.mid1pb;
   else                         rightpb=view->border.rightpb;
  }
  else                          rightpb=0;


/*  pagep=fob_find(v->page,v->file); */

  width=v->xpo+(v->file)->hdr.maxpagewidth[view->branch]+rightpb;

  err=bbc_rectanglefill(v->ox+osunits(width,view),v->oy,
                        osunits(v->xpe-width,view),-HMARGIN);
 }


 return(err);
}




static os_error * hrulermark(int a,int x,viewredrawstr * v,int hmargin)
{
//	cprintf("hmargin %d",hmargin);

 if(a)
 {
  x=v->ox+osunits(x+v->xpo,v->view);
  bbc_move(x,v->oy);
  return(bbc_draw(x,v->oy-hmargin));
 }

 return(NULL);
}



static os_error * redrawhrulersub(viewredrawstr * v,int more)
{
 os_error       * err;
 int              u;
 int              i;
 int              x;
 int              y;
 int              n;
 int              mz;
 int              measure;

 int              hstep;
 int              hbstep;
 int              ubstep;
 char    *        pattern;

 viewstr *        view;
 filestr *        file;

 int              ghost;
 pagestr        * page;
 sectionstr     * section;
 int              origin;
 int              hmargin;


 err=NULL;

 view=v->view;
 file=v->file;

 measure=view->data.units;

 hstep =hsteps[mz=mzone(view)][measure];
 hbstep=hbsteps[mz][measure];
 ubstep=ubsteps[mz][measure];
 pattern=hpattern[mz][measure];

 while(more)
 {
  v->ox=v->redrawstr.box.x0-v->redrawstr.scx;
  v->oy=v->redrawstr.box.y1-v->redrawstr.scy;

  hmargin=ypix2os(yos2pix(v->redrawstr.box.y1-v->redrawstr.box.y0)-1);

  if(!rmupdate)
  {
//   y=v->oy-ypix2os(yos2pix(HMARGIN));
			y=v->oy-hmargin;

   bbc_move(v->redrawstr.box.x0-cvdu.extrax,y);
   bbc_draw(v->redrawstr.box.x1+cvdu.extrax,y);
  }




  v->ylox=-(v->redrawstr.g.y1-v->oy)-cvdu.extray;
  v->yhix=-(v->redrawstr.g.y0-v->oy)+cvdu.extray;
  v->xlox=v->redrawstr.g.x0-v->ox-cvdu.extrax;
  v->xhix=v->redrawstr.g.x1-v->ox+cvdu.extrax;

  v->yloz=ourunits(v->ylox,view);
  v->yhiz=ourunits(v->yhix,view);
  v->xloz=ourunits(v->xlox,view);
  v->xhiz=ourunits(v->xhix,view);

  v->page=view->page;
  v->section=view->section;

  v->yps=view->ps;

  for(i=0;i<=view->data.dble;i++)
  {
   if(i==0) firstxcoord(v);
   else     nextxcoord(v);

   setxcoord(v);
   setycoord(v);

   if(v->xhi<v->xpe) v->xhi+=hstep*hbstep;

   origin=v->xpo+view->data.rulerx;
   n=(v->xlo-origin)/(hstep*hbstep);
   if(v->xlo<origin) n--;
   v->xlo=origin+n*hstep*hbstep;

   u=0;
   n*=ubstep;

   if(!rmupdate)
   {
    wimp_setcolour(WIMPBLACK);

    for(x=v->xlo;x<=v->xhi;x+=hstep)
    {
     if((x>=v->xps && x<=v->xpe))
     {
      /* using hmargin here because window may be bigger than HMARGIN */

      bbc_move(v->ox+osunits(x,view),v->oy-hmargin);
      bbc_draw(v->ox+osunits(x,view),
               v->oy-(hmargin-((pattern[u]-'0')*HMARGIN)/PATTERNSCALE));

      if(u==0)
      {
//     rulerprint(n,xcoordpix2os(xcoordos2pix(v->ox+osunits(x,view))-2),v->oy-1*(HMARGIN>>3),8+1);
       rulerprint(n,(v->ox+osunits(x,view))-cvdu.extrax,v->oy-1*(HMARGIN>>3),8+1);
       n+=ubstep;
      }
     }
     else 
      if(u==0) n+=ubstep;

     if(++u==hbstep) u=0;
    }
   }


   ghost=isghost(view,v->page,v->section,v->xps);

   if(rmfile==file && v->page==rmpage && !ghost)
   {
    seteorcol(WIMPWHITE,WIMPRED);

    if(rmupdate)
    {
     hrulermark(rmold.ax[0],rmold.x[0],v,hmargin);
     hrulermark(rmold.ax[1],rmold.x[1],v,hmargin);
    }

    hrulermark(rmcur.ax[0],rmcur.x[0],v,hmargin);
    hrulermark(rmcur.ax[1],rmcur.x[1],v,hmargin);
   }


   if(!ghost && !rmupdate) redrawhgaps(v,view);

   if(view->data.dble)
   {
    page=(pagestr*)fob_find(v->page,file);
    section=(sectionstr*)fob_find(v->section,file);

    if(!ghost && (page->chain.next || section->chain.next))
    {
     if(page->chain.next) v->page=page->chain.next;
     else
     {
      v->section=section->chain.next;
      section=(sectionstr*)fob_find(v->section,file);
      v->page=section->chain.child;
     }
    }
   }
  }

  wimp_get_rectangle(&v->redrawstr,&more);
 }


 return(err);
}





static os_error * redrawhruler(int handle,int userhandle)
{
 os_error      * err;
 viewredrawstr   v;
 int             more;

 initviewredraw(&v);

 v.redrawstr.w=handle;
 v.view=(viewstr*)userhandle;
 v.file=v.view->file;

 err=wimp_redraw_wind(&v.redrawstr,&more);
 if(!err)
 {
		if(more)
		{
   rulerfontstart();
   err=redrawhrulersub(&v,more);
   rulerfontend();
		}
 }

 return(err);
}



/*******************************************************************************************/


static os_error * redrawrulerpatchsub(viewredrawstr * v,int more)
{
 os_error       * err;
 int              x;
 int              y;


 err=NULL;


 while(more)
 {
  v->ox=v->redrawstr.box.x0-v->redrawstr.scx;
  v->oy=v->redrawstr.box.y1-v->redrawstr.scy;

//  y=v->oy-ypix2os(yos2pix(HMARGIN));
		y=v->oy-ypix2os(yos2pix(v->redrawstr.box.y1-v->redrawstr.box.y0)-1);

  bbc_move(v->redrawstr.box.x0-cvdu.extrax,y);
  bbc_draw(v->redrawstr.box.x1+cvdu.extrax,y);

//  x=v->ox+xpix2os(xos2pix(VMARGIN));
		x=v->ox+xpix2os(xos2pix(v->redrawstr.box.x1-v->redrawstr.box.x0)-1);

  bbc_move(x,v->redrawstr.box.y0-cvdu.extray);
  bbc_draw(x,v->redrawstr.box.y1+cvdu.extray);

  wimp_get_rectangle(&v->redrawstr,&more);
 }


 return(err);
}





static os_error * redrawrulerpatch(int handle,int userhandle)
{
 os_error      * err;
 viewredrawstr   v;
 int             more;

 initviewredraw(&v);

 v.redrawstr.w=handle;
 v.view=(viewstr*)userhandle;
 v.file=v.view->file;

 err=wimp_redraw_wind(&v.redrawstr,&more);
 if(!err)
 {
  err=redrawrulerpatchsub(&v,more);
 }

 return(err);
}




/*****************************************************************************/

static os_error * rulericon(int handle,int userhandle,wimp_mousestr * m,int h)
{
 os_error       * err;
 posnstr          posn;
 viewstr        * view;
 filestr        * file;
 wimp_mousestr    mx;
 windowstr        window;
 static int       lastobject;


 err=NULL;


 if(m->bbits==0x4 || m->bbits==0x40)
 {
  if(m->bbits==0x4) lastobject=0;

  view=(viewstr*)userhandle;
  file=view->file;

  if(!view->data.noclicks)
  {
   mx=*m;                   /* want to treat posn as edge of ruler */
   if(h)
   {
    getw(handle,&window);
    mx.y=window.y0;
   }

   initposn(&posn,view,file,&mx);

   if(!posn.shift && !posn.alt && (toolcurrent(file)==TL_SELECT))
   {
    err=getposn(&posn);
 
    if(!err && !posn.ghost)
    {
     if(m->bbits==0x4 && posn.valid)
     {
      if(!posn.ctrl)
      {
       err=rulercreate(&posn,h);
       lastobject=posn.object;
      }
      else  /* delete all vertical guidelines */
      {
       err=rulerdelete(file,h,posn.page,posn.section);
      }
     }
     else
     if(m->bbits==0x40 && lastobject && posn.valid)
     {
      err=undostart(file);
      if(!err)
      {
       posn.object=lastobject;
       clearselectionlo();
       objectselect((objhdrstr *)fob_find(lastobject,file),&posn,SELECTOBJECT);
       undoend(file);
       posn.otype=OGLINE;
       objectmaster(lastobject,file,&posn.master,&posn.bits);
       sdragselection(&posn,h?TOOLVLOCK:TOOLHLOCK);
      }
     }
    }
   }

   openinfopal(view,file);
   setactiveview(view);
  }
 }

 return(err);
}



static os_error * hrulericon(int handle,int userhandle,wimp_mousestr * m)
{
 return(rulericon(handle,userhandle,m,1));
}


static os_error * vrulericon(int handle,int userhandle,wimp_mousestr * m)
{
 return(rulericon(handle,userhandle,m,0));
}


/*****************************************************************************/

os_error * refreshrulers(viewstr * view)
{
 os_error * err;

 if(view->data.rulers)
 {
           err=refreshwindow(view->vruler);
  if(!err) err=refreshwindow(view->hruler);
 }
 else err=NULL;

 return(err);
}

os_error * closerulers(viewstr * view)
{
 os_error * err;

 if(view->data.rulers)
 {
  remredrawevent(redrawvruler,view->vruler,(int)view);
  remredrawevent(redrawhruler,view->hruler,(int)view);
  remredrawevent(redrawrulerpatch,view->rulerpatch,(int)view);
  remclickevent(vrulericon,view->vruler,(int)view);
  remclickevent(hrulericon,view->hruler,(int)view);

           err=closedown(&view->vruler);
  if(!err) err=closedown(&view->hruler);
  if(!err) err=closedown(&view->rulerpatch);

  viewmarginadjust(view,-view->vpshift,0,-view->hpshift,0);

  view->vpshift=0;
  view->hpshift=0;
 }
 else err=NULL;

 return(err);
}



// extra extent so that ruler always opens at correct scx,scy
// not forced to scroll by extents
//#define SCROLLEXTRA 10
// Plan B super big extents to cope with widow opens bigger than
// document workspace extent


os_error * setextentrulers(viewstr * view)
{
 os_error       * err;
 wimp_redrawstr   eblock;

 if(view->data.rulers && view->vruler && view->hruler)
 {
  eblock.w=view->vruler;
  eblock.box=view->extent;
  eblock.box.y0-=cvdu.screeny;
  eblock.box.y1+=cvdu.screeny;
  err=wimp_set_extent(&eblock);

  eblock.w=view->hruler;
  eblock.box=view->extent;
  eblock.box.x1+=cvdu.screenx;
  eblock.box.x0-=cvdu.screenx;
  err=wimp_set_extent(&eblock);
 }
 else err=NULL;

 return(err);
}



os_error * createrulers(viewstr * view)
{
 os_error * err;

 if(view->data.rulers)
 {
           err=createwindowparent(TVRULER,&view->vruler,view->handle);
  if(!err) err=createwindowparent(THRULER,&view->hruler,view->handle);
  if(!err) err=createwindowparent(TRULERPATCH,&view->rulerpatch,view->handle);
  addredrawevent(redrawvruler,view->vruler,(int)view);
  addredrawevent(redrawhruler,view->hruler,(int)view);
  addredrawevent(redrawrulerpatch,view->rulerpatch,(int)view);
  addclickevent(hrulericon,view->hruler,(int)view);
  addclickevent(vrulericon,view->vruler,(int)view);
  wimp_setwindowbackground(view->vruler,CCRGB+PALWHITE);
  wimp_setwindowbackground(view->hruler,CCRGB+PALWHITE);
  wimp_setwindowbackground(view->rulerpatch,CCRGB+PALWHITE);

  view->hpshift=HMARGIN;
  view->vpshift=VMARGIN;

  viewmarginadjust(view,view->vpshift,0,view->hpshift,0);
 }
 else err=NULL;

 return(err);
}



static os_error * addrulers(viewstr * view)
{
 os_error * err;

 if(!view->data.rulers)
 {
  view->data.rulers=1;

  err=createrulers(view);

  updateviewextent(view);

  viewopenshift(view,-VMARGIN,HMARGIN);
 }
 else err=NULL;

 return(err);
}



static os_error * remrulers(viewstr * view)
{
 os_error * err;

 if(view->data.rulers)
 {
  err=closerulers(view);
  view->data.rulers=0;

  updateviewextent(view);

  viewopenshift(view,VMARGIN,-HMARGIN);
 }
 else err=NULL;

 return(err);
}




os_error * openrulerpatch(viewstr * view,wimp_openstr * o)
{
 os_error   * err;
 wimp_openstr r;

 r.w=view->rulerpatch;
 r.x=0;
 r.y=0;
 r.behind=o->behind;

 r.box.x0=view->mainwindow.x0;
// r.box.x1=r.box.x0+VMARGIN;
 r.box.x1=xcoordpix2os(xcoordos2pix(r.box.x0)+xos2pix(VMARGIN)+1);

 r.box.y1=view->mainwindow.y1-view->hshift;
// r.box.y0=r.box.y1-HMARGIN;
 r.box.y0=ycoordpix2os(ycoordos2pix(r.box.y1)+yos2pix(HMARGIN)+1);


 err=wimp_open_wind(&r);

 return(err);
}



// we have to force rulers to be fixed pixel width
// but more difficult is to keep redraws of pixels in ruler
// in step with pixel in main window
// for example a guide line in the main window staying in
// the same position as the mark on the ruler
// we solve this by a fiddle, opening the rulers to the same
// dimensions as the main window.
// to do it the hard way one would have to ensure that
// the rulers scrolled by the same number of pixels as
// the main window does. There then has to be some calculation
// that links pixels in the two windows.




os_error * openvruler(viewstr * view,wimp_openstr * o)
{
 os_error   * err;
 wimp_openstr r;
 int          by;

 r.w=view->vruler;
 r.x=0;
 r.y=view->mainwindow.scy;//-view->hpshift-view->hshift; /* -HMARGIN; */ //mainpane
 r.behind=o->behind;

// have to force the ruler to be constant pixel width
 r.box.x0=view->mainwindow.x0+view->vshift;
 r.box.x1=xcoordpix2os(xcoordos2pix(r.box.x0)+xos2pix(VMARGIN)+1);

 r.box.y0=view->mainwindow.y0;
// r.box.y1=ycoordpix2os(ycoordos2pix(view->mainwindow.y1)+yos2pix(HMARGIN+view->hshift));
 r.box.y1=view->mainwindow.y1;

 by=r.box.y1-r.y;
// if((r.box.y0-by)<(view->extent.y0)) r.box.y0=view->extent.y0+by;


 err=wimp_open_wind(&r);

 return(err);
}




os_error * openhruler(viewstr * view,wimp_openstr * o)
{
 os_error   * err;
 wimp_openstr r;
 int          bx;


 r.w=view->hruler;
// r.x=xpix2os(xos2pix(view->mainwindow.scx)+xos2pix(view->vshift)+xos2pix(view->vpshift));//+cvdu.deltax; //mainpane
 r.x=view->mainwindow.scx;

 r.y=0;
 r.behind=o->behind;

// r.box.x0=xpix2os(xos2pix(view->mainwindow.x0)+xos2pix(VMARGIN)+xos2pix(view->vshift));

 r.box.x0=view->mainwindow.x0;
 r.box.x1=view->mainwindow.x1;

 // clipping, stopping ruler being opened bigger than extent, stop scroll

 bx=r.box.x0-r.x;
// if((r.box.x1-bx)>(view->extent.x1)) r.box.x1=view->extent.x1+bx;

 r.box.y1=o->box.y1-view->hshift;
 r.box.y0=ycoordpix2os(ycoordos2pix(r.box.y1)+yos2pix(HMARGIN)+1);

 err=wimp_open_wind(&r);

 return(err);
}




int getrulers(int fp)
{
 return(menuview->data.rulers);
 USE(fp);
}




void xsetrulers(int open,viewstr * view)
{
 filestr * file;

 file=view->file;

 if(open && !view->data.rulers) addrulers(view);
 else
 if(!open && view->data.rulers) remrulers(view);
}


void setrulers(int fp)
{
 xsetrulers(stack[fp],menuview);
 updateviewdata(menuview);
}


void togglerulers(viewstr * view)
{
 xsetrulers(view->data.rulers^1,view);
 updateviewdata(view);
}

