/*->c.toolbox */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <math.h>  
#include <limits.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "bbc.h"
#include "akbd.h"
#include "transform.h"

#include "err.h"
#include "fsx.h"
#include "bf.h"
#include "xext.h"
#include "poll.h"
#include "alloc.h"
#include "trans.h"
#include "etc.h"
#include "colour.h"
#include "key.h"
#include "xmath.h"
#include "dbhi.h"
#include "deb.h"

#include "str.h"
#include "constants.h"

#include "mhi.h"
#include "object.h"
#include "chain.h"

#include "reslink.h"
#include "viewo.h"
#include "viewr.h"
#include "view.h"
#include "file.h"
#include "caret.h"
#include "select.h"
#include "selaction.h"
#include "posn.h"
#include "xredraw.h"
#include "curve.h"
#include "undo.h"
#include "frameflow.h"
#include "geometry.h"
#include "units.h"
#include "xcolour.h"
#include "objectdb.h"
#include "grid.h"
#include "pref.h"
#include "gline.h"
#include "rulers.h"



#include "infopal.h"
#include "editpath.h"
#include "toolstr.h"
#include "tools.h"
#include "frametool.h"
#include "toolbox.h"



#define UCREATES 0x9000
#define UCREATEF 0x9001




static pointstr resizeptr=
{
 "p_resize",2,2
};


static pointstr moveptr=
{
 "p_move",10,10
};


static pointstr vertglptr=
{
 "p_vertgl",11,7
};


static pointstr horizglptr=
{
 "p_horizgl",7,11
};



static pointstr rotateptr=
{
 "p_rotate",10,11
};


static pointstr lockptr=
{
 "p_lock",10,10
};


static pointstr magnifyptr=
{
 "p_magnify",11,11
};



static pointstr ddptr=
{
 "p_export",0,0
};




/*****************************************************************************/

static int          magstate;           /* from magnify tool */

static os_error * magdragend(toolboxstr * toolbox,posnstr * posn,
                                                       wimp_box * box);

static os_error * magcancel(toolboxstr * toolbox,viewstr * view,filestr* file);

static os_error * magmove(toolboxstr * toolbox,posnstr * posn,int move);

static os_error * magredraw(toolboxstr * toolbox,viewredrawstr * v);

static os_error * magclick(toolboxstr * toolbox,posnstr * posn);






/* Select tool */


static int          selectstate;        /* used to control redraws    */
static int          selecttranstype;    /* used to control transforms */
static posnstr      sstartposn;
static transformstr stransform;
static int          snegativescale;
static int          selectaltstate;     /* alt drag box */
static int          selectddstate;      /* drag n drop  */
static posnstr      scurrposn;
static int          sadjust;









#define STRANSTOPLEFT  1 /* top left        */
#define STRANSTOPMID   2 /* top mid         */
#define STRANSTOPRIGHT 3 /* top right       */
#define STRANSRIGHTMID 4 /* right mid       */
#define STRANSBOTRIGHT 5 /* bot right       */
#define STRANSBOTMID   6 /* bot mid         */
#define STRANSBOTLEFT  7 /* bot left        */
#define STRANSLEFTMID  8 /* left mid        */
#define STRANSSOLID    9 /* shift the lot   */
#define STRANSHORIZ   10 /* left/right only */
#define STRANSVERT    11 /* up/down only    */
#define STRANSLOCK    12 /* locked          */




static os_error * ssetptr(int boxhandle)
{
 if(selectstate==STRANSHORIZ) toolschangepointer(128,USERPOINTER,&vertglptr);
 else
 if(selectstate==STRANSVERT)  toolschangepointer(128,USERPOINTER,&horizglptr);
 else
 if(selectstate==STRANSLOCK)  toolschangepointer(128,USERPOINTER,&lockptr);
 else
 if(selectstate==STRANSSOLID) toolschangepointer(128,USERPOINTER,&moveptr);
 else
 if(selectstate)              toolschangepointer(128,USERPOINTER,&resizeptr);
 else
 if(selectaltstate)           toolschangepointer(128,SYSTEMPOINTER,NULL);
 else
 if(selectddstate)            toolschangepointer(128,USERPOINTER,&ddptr);
 else
 if(boxhandle)                toolschangepointer(8,USERPOINTER,&resizeptr);
 else                         toolschangepointer(0,SYSTEMPOINTER,NULL);

 return(NULL);
}



static int strans(transformstr * r,transformstr * t,
                                    curvestr * c,transformstr * ct)
{
 transformstr    temp1;
 transformstr    temp2;
 transformstr    temp3;
 int             xshift;
 int             yshift;
 int             fix;
 int             x0;
 int             x1;
 int             y0;
 int             y1;


 fix=0;

 if(selecttranstype>=STRANSTOPLEFT && selecttranstype<STRANSSOLID)
 {
  xshift=yshift=0;

  x0=c->box.x;
  x1=x0+c->box.w;
  y0=c->box.y;
  y1=y0+c->box.h;


  switch(selecttranstype)
  {
   case  STRANSTOPLEFT: /* top left */
                       fix|=SFIXRIGHT;
   case   STRANSTOPMID: /* top mid  */
                       fix|=SFIXBOTTOM;
                       xshift=x1;
                       yshift=y1;
                       break;

   case STRANSTOPRIGHT: /* top right */
                       fix|=SFIXBOTTOM;
   case STRANSRIGHTMID: /* right mid */
                       xshift=x0;
                       yshift=y1;
                       break;

   case STRANSBOTRIGHT: /* bot right */
   case   STRANSBOTMID: /* bot mid   */
                       xshift=x0;
                       yshift=y0;
                       break;

    case STRANSBOTLEFT: /* bot left */
    case STRANSLEFTMID: /* left mid */
                       fix|=SFIXRIGHT;
                       xshift=x1;
                       yshift=y0;
                       break;
  }

  tr_setshift(&temp1,-xshift,-yshift);
  tr_multiply(&temp2,&temp1,&stransform);
  tr_setshift(&temp1,xshift,yshift);
  tr_multiply(&temp3,&temp2,&temp1);
  tr_multiply(r,&temp3,t);

  return(fix|TR_SCALE);
 }
 else
 if(selecttranstype==STRANSSOLID || 
    selecttranstype==STRANSHORIZ || selecttranstype==STRANSVERT)
 {
  tr_multiply(r,t,&stransform);
  return(TR_SHIFT);
 }

 return(0);

 USE(ct);
}



/*

  was

  goes wrong if ruler origins not equal

 if(sstartposn.view->data.grid.lock)
 {
  if(type & (TR_PLUS90|TR_MINUS90)) SWOP(point.x,point.y);

  if(bits & (1|2)) point.x=gridsnapx(point.x,sstartposn.view,0);
  if(bits & (4|8)) point.y=gridsnapy(point.y,sstartposn.view,0);

  if(type & (TR_PLUS90|TR_MINUS90)) SWOP(point.x,point.y);
 }

*/



static void ssnapfn(int * shiftx,int * shifty,int x,int y,
                                         transformstr * tr,int bits)
{
 coordstr point;
 int      type;
 int      xbits;

 point.x=x+(*shiftx);
 point.y=y+(*shifty);

 tr_transformset(&point,1,tr);

 type=tr_type(tr);


 xbits=bits;

 if(sstartposn.view->data.grid.lock)
 {
  if(sstartposn.view->data.grid.lock & GRIDLOCK)
  {
   if(bits & (1|2)) point.x=gridsnapx(point.x,sstartposn.view,0);
   xbits&=~(1|2);
  }

  if(bits & (4|8)) point.y=gridsnapy(point.y,sstartposn.view,0);
  xbits&=~(4|8);
 }

 if(!(sstartposn.view->data.grid.lock & GRIDLOCK)) 
 {
  snap((xbits & 0x1)?&point.x:NULL,
       (xbits & 0x2)?&point.x:NULL,
       (xbits & 0x4)?&point.y:NULL,
       (xbits & 0x8)?&point.y:NULL,
       &sstartposn,NULL);
 }

 tr_inverseset(&point,1,tr);

 if(bits & 1) *shiftx=point.x-x;
 else
 if(bits & 2) *shiftx=x-point.x;

 if(bits & 4) *shifty=point.y-y;
 else
 if(bits & 8) *shifty=y-point.y;
}



static void setstransform(transformstr * stransform,posnstr * posn)
{
 sobjectstr    * sob;
 coordstr        cpoint;
 coordstr        spoint;
 int             shiftx=0;
 int             shifty=0;
 int             w;
 int             h;
 int             neww;
 int             newh;
 viewstr       * view;
 filestr       * file;
 int             xl;
 int             xr;
 int             yb;
 int             yt;
 transformbitstr b;
 int             slock;
 outlineinfostr  oinfo;
 boxstr          box;


 view=posn->view;
 file=posn->file;

 snegativescale=0;

 if(selecttranstype>=STRANSTOPLEFT && selecttranstype<STRANSSOLID)
 {
  sob=(sobjectstr*)fob_find(sstartposn.object,posn->file);
  w=sob->fcurve.box.w;
  h=sob->fcurve.box.h;
  xl=sob->fcurve.box.x;
  yt=sob->fcurve.box.y;
  xr=xl+w;
  yb=yt+h;



  cpoint.x=posn->x;
  cpoint.y=posn->y;
  spoint.x=sstartposn.x;
  spoint.y=sstartposn.y;

  tr_inverseset(&cpoint,1,&sob->transform);
  tr_inverseset(&spoint,1,&sob->transform);

  shiftx=cpoint.x-spoint.x;
  shifty=cpoint.y-spoint.y;
  slock=1;

  /* convert posn into object space */
  /* work out transform             */

  switch(selecttranstype)
  {
   case  STRANSTOPLEFT:   /* top left */
                       ssnapfn(&shiftx,&shifty,xl,yt,&sob->transform,1|4);
                       break;

   case   STRANSTOPMID:  /* top middle */
                       slock=shiftx=0;
                       ssnapfn(&shiftx,&shifty,(xl+xr)/2,yt,&sob->transform,4);
                       break;

   case STRANSTOPRIGHT:  /* top right */
                       ssnapfn(&shiftx,&shifty,xr,yt,&sob->transform,2|4);
                       break;

   case STRANSRIGHTMID:  /* middle right */
                       slock=shifty=0;
                       ssnapfn(&shiftx,&shifty,xr,(yt+yb)/2,&sob->transform,2);
                       break;

   case STRANSBOTRIGHT: /* bot right */
                       ssnapfn(&shiftx,&shifty,xr,yb,&sob->transform,2|8);
                       break;

   case   STRANSBOTMID: /* bot middle */
                       slock=shiftx=0;
                       ssnapfn(&shiftx,&shifty,(xl+xr)/2,yb,&sob->transform,8);
                       break;

   case  STRANSBOTLEFT: /* bot left */
                       ssnapfn(&shiftx,&shifty,xl,yb,&sob->transform,1|8);
                       break;

   case  STRANSLEFTMID: /* middle left */
                       slock=shifty=0;
                       ssnapfn(&shiftx,&shifty,xl,(yt+yb)/2,&sob->transform,1);
                       break;
  }

  neww=w-shiftx;
  newh=h-shifty;

  if(sstartposn.shift && slock)
  {
   tr_getbits(&sob->transform,&b);

   if(neww>newh) newh=scale(neww,b.xscale,b.yscale);
   else          neww=scale(newh,b.yscale,b.xscale);
  }
  else
  if(sstartposn.ctrl && slock)
  {
   if(neww>newh) newh=scale(h,neww,w);
   else          neww=scale(w,newh,h);
  }

  snegativescale=(neww<0)||(newh<0);

  tr_setscale(stransform,w,neww,h,newh);
 }
 else
 {
  {
   if(view->data.grid.lock)
   {
    cpoint.x=gridsnapx(posn->x,view,0);
    cpoint.y=gridsnapy(posn->y,view,0);
    spoint.x=gridsnapx(sstartposn.x,view,0);
    spoint.y=gridsnapy(sstartposn.y,view,0);
    shiftx=cpoint.x-spoint.x;
    shifty=cpoint.y-spoint.y;
   }

   if(!(view->data.grid.lock & GRIDLOCK))
   {
    objectoread(SELECTINFO,sstartposn.object,sstartposn.file,&oinfo);
    getmaxbox(&box,&oinfo.fcurve.box,&oinfo.tr); /* NB Dnd file above */

    xl=box.x;
    xr=xl+box.w;
    yt=box.y;
    yb=yt+box.h;

    xl+=posn->x-sstartposn.x;
    xr+=posn->x-sstartposn.x;
    yt+=posn->y-sstartposn.y;
    yb+=posn->y-sstartposn.y;

    snap(&xl,&xr,&yt,&yb,&sstartposn,selectddstate?posn:NULL); 
                                                           /* uses object */
    shiftx=xl-box.x;
    if(!(view->data.grid.lock & GRIDBASELOCK)) shifty=yt-box.y;
   }

   switch(selecttranstype)
   {
    case  STRANSLOCK:
                     shiftx=0;
                     shifty=0;
                     break;

    case STRANSHORIZ:
                     shifty=0;
                     break;

    case  STRANSVERT:
                     shiftx=0;
                     break;

    case STRANSSOLID:
                     if(sstartposn.shift && !selectddstate)
                     {
                      if(ABS(shiftx)>ABS(shifty)) shifty=0;
                      else                        shiftx=0;
                     }
                     break;
   }
  }

  tr_setshift(stransform,shiftx,shifty);
 }
}



static os_error * sredraw(toolboxstr * toolbox,viewredrawstr * v)
{
 viewredrawstr v2;

 initviewredraw(&v2);


 if(magstate) return(magredraw(toolbox,v));


 if(selectaltstate && sstartposn.page==v->page)
 {
  rseteorcol(WIMPWHITE,WIMPGREY2);
  bbc_rectangle(
           v->ox+osunits(sstartposn.xpo+sstartposn.x,v->view),
           v->oy-osunits(sstartposn.ypo+sstartposn.y,v->view),
               osunits(scurrposn.x-sstartposn.x,v->view),
               osunits(sstartposn.y-scurrposn.y,v->view));
 }
 else
 if(selectddstate && scurrposn.page==v->page)
 {
  v2=*v;
  v2.file=sstartposn.file;
  redrawselectiont(&v2,strans,0);
 }
 else
 if(selectstate && selectstate!=STRANSLOCK && sstartposn.page==v->page)
 {
  redrawselectiont(v,strans,1);
 }
 return(NULL);
 USE(toolbox);
}



static int findboxp(posnstr * posn,int before)
{
 int code;

 if(before) before=getfirstselected(posn->file,posn->page);
 else       before=-1;

 if((code=findbox(posn,before))==0 && before>=0)
 {
  return(findbox(posn,-1));
 }

 return(code);
}




static int canmoveobject(posnstr * posn)
{
 filestr    * file;
 int          object;
 int          master;
 objectbitstr bits;

 if((!posn->master || posn->bits.local) && !posn->bits.locked)
 {
  file=posn->file;

  if(!embededobject(posn->object,file))
  {
   object=selectedparent(posn->object,file);

   if(object==posn->object) return(1);

   if(object && !embededobject(object,file))
   {
    objectmaster(object,file,&master,&bits);
    if((!master || bits.local) && !bits.locked)
    {
     return(1);
    }
   }
  }
 }
 return(0);
}




static os_error * srulermarks(int op)
{
 os_error     * err;
 curvestr       xcurve;
 transformstr   r;
 sobjectstr   * sob;
 rmstr          rmarks;
 boxstr         maxbox;

 if(op==1 || op==2)
 {
  sob=(sobjectstr*)fob_find(sstartposn.object,sstartposn.file);
  xcurve=sob->fcurve;
  strans(&r,&sob->transform,&sob->fcurve,&sob->transform); 
  correctcurve(&r,&sob->transform,&xcurve);

  if(sstartposn.bits.hlock) r.e=sob->transform.e;
  else
  if(sstartposn.bits.vlock) r.f=sob->transform.f;

  getmaxbox(&maxbox,&xcurve.box,&r);

  rmarks.ax[0]=rmarks.ay[0]=1;
  rmarks.ax[1]=(maxbox.w!=0);
  rmarks.ay[1]=(maxbox.h!=0);

  rmarks.x[0]=maxbox.x;
  rmarks.y[0]=maxbox.y;
  rmarks.x[1]=maxbox.x+maxbox.w;
  rmarks.y[1]=maxbox.y+maxbox.h;


  if(op==1)
  {
   err=rulermarkstart(&rmarks,sstartposn.page,
                              sstartposn.branch,sstartposn.file);
  }
  else
  {
   err=rulermarkupdate(&rmarks);
  }
 }
 else
 {
  err=rulermarkend();
 }

 return(err);
}



/* trying to ensure that always leave this function with posn */
/* object etc. set up the same way                            */

static os_error * sclicks(toolboxstr * toolbox,posnstr * posn,int ptr)
{
 int boxhandle;
 int object;
 int select;
 int adjust;

 boxhandle=0;

// cprintf("buttons %x",posn->buttons);


 if(posn->buttons & (VBCLICKS|VBCLICKA))
 {
  sstartposn=*posn;
//  cprintf("shift %d",sstartposn.shift);
 }
 else
 if(posn->buttons & VBLONGCLICKS)
 {
 // need to set startposn on longclicks to cope with SHIFT+longclick starting
 // a constrained drag.
 // shift+click is grabbed by the text tool and never gets here

  findbox(posn,-1);
  sstartposn=*posn;
  cprintf("long click");
 }


 if(!posn->ghost)
 {
  select= posn->buttons==VBLONGCLICKS || 
         (posn->buttons==(VBLONGCLICKS|VBCTRL))  ||
         (posn->buttons==(VBLONGCLICKS|VBSHIFT)) ||
          posn->buttons==VBDRAGCS     || 
          posn->buttons==VBDRAGSS     ||
          posn->buttons==VBDRAGS;

  adjust= posn->buttons==VBLONGCLICKA ||
         (posn->buttons==(VBLONGCLICKA|VBCTRL))  ||
         (posn->buttons==(VBLONGCLICKA|VBSHIFT)) ||
          posn->buttons==VBDRAGCA     ||
          posn->buttons==VBDRAGSA     ||
          posn->buttons==VBDRAGA;

  if((select || adjust) && findboxhandle(&sstartposn))
  {
   if((!sstartposn.master || sstartposn.bits.local) && !sstartposn.bits.locked)
   {
    sadjust=adjust;
    selecttranstype=selectstate=sstartposn.handle;
   }
  }
  else
  if(findboxhandle(posn))
  {
   boxhandle=1;
  }
  else
  {
   if(findboxp(posn,posn->buttons==(VBDBLCLICKS|(VBCTRL|VBSHIFT))))
   {
    if(posn->buttons & (VBCLICKS|VBCLICKA)) sstartposn=*posn;
   /* need this here to get object */


    if(posn->buttons==VBCLICKS                 || 
      /* posn->buttons==VBCLICKCS                ||  stop control+select selecting things */
       posn->buttons==VBCLICKCA                ||   // allow adjust+control select
       posn->buttons==VBCLICKSS                ||
       posn->buttons==(VBCLICKS|VBALT)         ||
       posn->buttons==(VBCLICKS|VBALT|VBSHIFT) ||
       posn->buttons==(VBDBLCLICKS|(VBCTRL|VBSHIFT)))
    {
     if(!isselectedbig(posn->object,posn->file))
     {
      changeselection(posn,SELECTOBJECT,1);
     }
    }
    else
    if(posn->buttons==VBCLICKA || posn->buttons==VBCLICKCS) /* single click adjust */  /* allow control+adjust select */
    {
     if((object=isselectedbig(posn->object,posn->file))!=0)
     {
      posn->object=object;
      changeselection(posn,DESELECTOBJECT,0);
     }
     else
     {
      changeselection(posn,SELECTOBJECT,0);
     }
    }
    else
    if(posn->buttons==VBLONGCLICKS ||
       posn->buttons==(VBLONGCLICKS|VBSHIFT) ||
       posn->buttons==(VBLONGCLICKS|VBCTRL) ||
    /*   posn->buttons==VBDRAGCS || */
       posn->buttons==VBDRAGCA ||
       posn->buttons==(VBDRAGA|VBCTRL|VBSHIFT) ||  // new 
       posn->buttons==(VBLONGCLICKA |VBCTRL) ||
       posn->buttons==(VBLONGCLICKA|VBCTRL|VBSHIFT) || // new
       posn->buttons==VBDRAGSS)
    {
     if(canmoveobject(posn))
     {
      if(posn->bits.vlock) selecttranstype=selectstate=STRANSHORIZ;
      else
      if(posn->bits.hlock) selecttranstype=selectstate=STRANSVERT;
      else                 selecttranstype=selectstate=STRANSSOLID;
     }
     else
     {
      selecttranstype=selectstate=STRANSLOCK;
     }
    }
    else
    if(posn->buttons==(VBDRAGS|(VBCTRL|VBSHIFT)))
    {
     selectaltstate=1;
    }
    else
    if(posn->buttons==(VBDRAGA|(VBCTRL|VBSHIFT)))
    {
     selectaltstate=2;
    }
    else
    if(posn->buttons==(VBDRAGS|VBALT) || 
       posn->buttons==(VBLONGCLICKS|VBALT))
    {
     selecttranstype=STRANSSOLID;
     selectddstate=1;
    }
    else
    if(posn->buttons==(VBDRAGS|VBALT|VBSHIFT) || 
       posn->buttons==(VBLONGCLICKS|VBALT|VBSHIFT))
    {
     selecttranstype=STRANSSOLID;
     selectddstate=2;
    }
   }
   else  /*  not on a box or anything */
   if(!posn->ghost)
   {
    if(posn->buttons==VBCLICKS  || 
       posn->buttons==VBCLICKCS || posn->buttons==VBCLICKSS)
    {
     clearselection();
    }
    else
    if(posn->buttons==(VBDRAGS|(VBCTRL|VBSHIFT)))
    {
     selectaltstate=1;
    }
    else
    if(posn->buttons==(VBDRAGA|(VBCTRL|VBSHIFT)))
    {
     selectaltstate=2;
    }
   }
  }

  if(ptr) ssetptr(boxhandle);

  if(selectaltstate)
  {
   scurrposn=*posn;
   toolsstartdrag(posn,TOOLDRAGPAGE|TOOLDRAGKN|TOOLDRAGPTR);
  }

  if(selectddstate)
  {
   scurrposn=*posn;
   setstransform(&stransform,&sstartposn); /* need selectstate set */
   toolsstartdrag(posn,TOOLDRAGTRANSFER|TOOLDRAGPTR);
   viewrefreshall(posn->file,posn->view->branch,VRTOOLS|VRONELAYER);
  }

  if(selectstate)
  {
   setstransform(&stransform,&sstartposn); /* need selectstate set */
   toolsstartdrag(posn,TOOLDRAGPAGE|
                       TOOLDRAGKN|
                       TOOLDRAGPTR |
                       ((selectstate==STRANSLOCK)?TOOLDRAGNS:0));
   if(selectstate!=STRANSLOCK) 
              viewrefreshall(posn->file,posn->view->branch,VRTOOLS|VRONELAYER);
   srulermarks(1);
  }

 }

 caretensure(posn->view);

 return(NULL);
 USE(toolbox);
}



static os_error * sclick(toolboxstr * toolbox,posnstr * posn)
{
 if((posn->buttons & (VBCTRL|VBALT))==(VBCTRL|VBALT)) 
                                             return(magclick(toolbox,posn));

 else                                        return(sclicks(toolbox,posn,1));
}



os_error * sdragselection(posnstr * posn,int bits)
{
 sstartposn=*posn;
 selecttranstype=selectstate=STRANSSOLID;
 setstransform(&stransform,posn); /* need selectstate set */
 toolsstartdrag(posn,TOOLDRAGPAGE|TOOLDRAGKN|TOOLDRAGPTR|TOOLFORCE|bits);
 viewrefreshall(posn->file,posn->view->branch,VRTOOLS|VRONELAYER);
 return(srulermarks(1));
}




static os_error * sdragbox(void)
{
 pagestr     * pagep;
 int           child;
 sobjectstr  * link;
 boxstr        box;

 if(sstartposn.x<scurrposn.x)
 {
  box.x=sstartposn.x;
  box.w=scurrposn.x-sstartposn.x;
 }
 else
 {
  box.x=scurrposn.x;
  box.w=sstartposn.x-scurrposn.x;
 }

 if(sstartposn.y<scurrposn.y)
 {
  box.y=sstartposn.y;
  box.h=scurrposn.y-sstartposn.y;
 }
 else
 {
  box.y=scurrposn.y;
  box.h=sstartposn.y-scurrposn.y;
 }

 pagep=(pagestr*)fob_find(sstartposn.page,sstartposn.file);
 child=pagep->chain.child;
 while(child)
 {
  link=(sobjectstr*)fob_find(child,sstartposn.file);
  sstartposn.object=child;
  child=link->chain.next;

  if(boxinbox(&link->box,&box))
  {
  /* objectselect((objhdrstr *)link,&sstartposn,SELECTOBJECT); */
   changeselection(&sstartposn,SELECTOBJECT,0);
  }
 }

 return(NULL);
}



static os_error * sdragend(toolboxstr * toolbox,posnstr * posn,wimp_box * box)
{
 os_error * err;
 aboxstr    abox;
 int        oldstate;

 if(magstate) return(magdragend(toolbox,posn,box));


 err=NULL;

 if(selectaltstate)
 {
  abox.x0=sstartposn.xps-sstartposn.xpo;
  abox.x1=sstartposn.xpe-sstartposn.xpo;
  abox.y1=sstartposn.ype-sstartposn.ypo;
  abox.y0=sstartposn.yps-sstartposn.ypo;

  viewrefresharea(posn->file,posn->page,&abox,VRTOOLS|VRONELAYER);
  oldstate=selectaltstate;
  selectaltstate=0;

  undostart(posn->file);
  if(oldstate==1) clearselection();
  err=sdragbox();
  undoend(posn->file);
 }
 else
 if(selectddstate)
 {
  oldstate=selectddstate;

  abox.x0=scurrposn.xps-scurrposn.xpo;
  abox.x1=scurrposn.xpe-scurrposn.xpo;
  abox.y1=scurrposn.ype-scurrposn.ypo;
  abox.y0=scurrposn.yps-scurrposn.ypo;

  viewrefresharea(scurrposn.file,scurrposn.page,&abox,VRTOOLS|VRONELAYER);

  selectddstate=0;

  if(posn->valid && !posn->ghost && !(posn->view)->data.nodnd)
  {
   if(posn->file!=sstartposn.file) oldstate=3-oldstate;
   findbox(posn,-1);
   err=moveselection(&sstartposn,posn,oldstate==2,strans,&stransform);
  }
 }
 else
 if(selectstate)
 {
  abox.x0=sstartposn.xps-sstartposn.xpo;  /* NB startposn not curr as above */
  abox.x1=sstartposn.xpe-sstartposn.xpo;
  abox.y1=sstartposn.ype-sstartposn.ypo;
  abox.y0=sstartposn.yps-sstartposn.ypo;

  if(selectstate!=STRANSLOCK)
  {
   viewrefresharea(posn->file,posn->page,&abox,VRTOOLS|VRONELAYER);
  }

  selectstate=0;                              /* stop redraw         */
  if(selecttranstype!=STRANSLOCK)
  {
   err=transformselection(strans,(sadjust?TRADJUST:0)|
                                 (snegativescale?TRMIRROR:0));
  }
  selecttranstype=0;                          /* but allow transform */
  iprefreshpane(posn->file,0x1,IP_PANE1);
  srulermarks(0);
 }

 return(err);

 USE(toolbox);
 USE(posn);
 USE(box);
}



static os_error * smove(toolboxstr * toolbox,posnstr * posn,int move)
{
 aboxstr      abox;
 transformstr newtransform;


 if(magstate) return(magmove(toolbox,posn,move));


 if(move)
 {
  if((selectaltstate || selectstate) && !posn->ghost)
  {

   if(posn->page==sstartposn.page && posn->file==sstartposn.file && 
      selectstate!=STRANSLOCK)
   {
    abox.x0=sstartposn.xps-sstartposn.xpo;
    abox.x1=sstartposn.xpe-sstartposn.xpo;
    abox.y1=sstartposn.ype-sstartposn.ypo;
    abox.y0=sstartposn.yps-sstartposn.ypo;

    if(selectaltstate)
    {
     viewrefresharea(posn->file,posn->page,&abox,VRTOOLS|VRONELAYER);
     scurrposn=*posn;
     viewrefresharea(posn->file,posn->page,&abox,VRTOOLS|VRONELAYER);
    }
    else
    if(selectstate && selectstate!=STRANSLOCK)
    {
     setstransform(&newtransform,posn);

     if(tr_cmp(&newtransform,&stransform))
     {
      viewrefresharea(posn->file,posn->page,&abox,VRTOOLS|VRONELAYER);
      stransform=newtransform;
      viewrefresharea(posn->file,posn->page,&abox,VRTOOLS|VRONELAYER);

      srulermarks(2);
     }
     iprefreshpane(posn->file,0x1,IP_PANE1);
    }
   }
  }
  else
  if(selectddstate)
  {
   if(!posn->ghost)
   {
    abox.x0=scurrposn.xps-scurrposn.xpo;
    abox.x1=scurrposn.xpe-scurrposn.xpo;
    abox.y1=scurrposn.ype-scurrposn.ypo;
    abox.y0=scurrposn.yps-scurrposn.ypo;

    setstransform(&newtransform,posn);
    if(tr_cmp(&newtransform,&stransform))
    {
     viewrefresharea(scurrposn.file,scurrposn.page,&abox,VRTOOLS|VRONELAYER);
     scurrposn=*posn;
     stransform=newtransform;

     abox.x0=posn->xps-posn->xpo;
     abox.x1=posn->xpe-posn->xpo;
     abox.y1=posn->ype-posn->ypo;
     abox.y0=posn->yps-posn->ypo;

     viewrefresharea(posn->file,posn->page,&abox,VRTOOLS|VRONELAYER);
    }
    iprefreshpane(posn->file,0x1,IP_PANE1);
   }
  }
  else
  {
   if(!posn->ghost && findboxhandle(posn)) 
                           toolschangepointer(8,USERPOINTER,&resizeptr);
   else                    toolschangepointer(0,SYSTEMPOINTER,NULL);
  }
 }

 return(NULL);
 USE(toolbox);
}



#define SIPX 0
#define SIPY 1
#define SIPH 3
#define SIPW 2


static iconmap sipmap[]=
{
 9,0,0,IP_TIPS,
 3,0,0,IP_TIPS,
 5,0,0,IP_TIPS,
 7,0,0,IP_TIPS,


 0,0,0,0,      /* terminator */
};



/*

"Select tool"     Select tool with no active object.
"Text frame"      Select tool with single object.
"n text frames"   Select tool with multiple objects.
"n objects"       Select tool with mixed objects.

*/


static os_error * siprefresh(viewstr * view,int handle,int code)
{
 int                object;
 sobjectstr       * sob;
 transformstr       tr;
 filestr          * file;
 int                n;
 objectdbstr        cont;
 char               string[64];
 boxbitstr          bits;


 file=view->file;

 n=nselected(&object,file);

 if(code & 0x2) 
 {
  if(!n) writeiconf(view->info,IP_HELPI,"%s","{SEL00}");
  else
  {
   cont.name=string;

   if(n==1)
   {
    cont.type=ODB_OBJECTNAME;
    sob=(sobjectstr*)fob_find(object,file);
    objectdb((objhdrstr*)sob,file,view,&cont);
    writeiconf(view->info,IP_HELPI,"%s",cont.name);
   }
   else
   if(selectiontype(&object,file)>0)
   {
    cont.type=ODB_OBJECTNAMEPLURAL;
    sob=(sobjectstr*)fob_find(object,file);
    objectdb((objhdrstr*)sob,file,view,&cont);
    writeiconf(view->info,IP_HELPI,"%d %s",n,cont.name);
   }
   else
   {
    writeiconf(view->info,IP_HELPI,"%d {SEL01}",n);
   }
  }
 }

 if(code & 0x3)
 {
  if(n)
  {
   sob=(sobjectstr*)fob_find(object,file);

   if(selectstate)
   {
    strans(&tr,&sob->transform,&sob->fcurve,&sob->transform);
    getboxbits(&sob->fcurve.box,&tr,&bits);
   }
   else
   {
    getboxbits(&sob->fcurve.box,&sob->transform,&bits);
   }

   writeiconc(handle,sipmap[SIPX].realicon,
                     unitstostring(bits.orgx-prefp(file)->rulerx,
                     view->data.units));

   writeiconc(handle,sipmap[SIPY].realicon,
                     unitstostring(bits.orgy-prefp(file)->rulery,
                     view->data.units));

   writeiconc(handle,sipmap[SIPW].realicon,
                      unitstostring(bits.width,view->data.units));
   writeiconc(handle,sipmap[SIPH].realicon,
                      unitstostring(bits.height,view->data.units));
  }
  else
  {
   writeiconc(handle,sipmap[SIPX].realicon,"");
   writeiconc(handle,sipmap[SIPY].realicon,"");
   writeiconc(handle,sipmap[SIPW].realicon,"");
   writeiconc(handle,sipmap[SIPH].realicon,"");
  }
 }

 return(NULL);
}








static boxbitstr * obp;
static boxbitstr * nbp;
static boxbitstr * nbshiftp;


static os_error * sipmodfn(int object,filestr * file)
{
 int          master;
 objectbitstr bits;

 objectmaster(object,file,&master,&bits);

 if(bits.noresize) setboxtrans(obp,nbshiftp);
 else              setboxtrans(obp,nbp);

 return(NULL);
}


static int sxsnap(viewstr * view,int posn)
{
 int mul;
 int div;

 switch(view->data.units)
 {
  case  INCHES:
               mul=(72000/20);
               div=1;
               break;

  case  POINTS:
               mul=(3*1000);
               div=1;
               break;

  case  MILIMS:
       default:
               mul=360000;
               div=127;
               break;
 }

 posn=scale(posn,div,mul);
 posn=scale(posn,mul,div);

 return(posn);
}


static os_error * sipaction(viewstr * view,int handle,int step,int index)
{
 os_error   * err;
 filestr    * file;
 int          n;
 int          object;
 boxbitstr    ob;
 boxbitstr    nb;
 boxbitstr    nbshift;
 char       * p;
 sobjectstr * sob;

 err=NULL;


 file=view->file;

 n=nselected(&object,file);
 if(n)
 {
  sob=(sobjectstr*)fob_find(object,file);

  getboxbits(&sob->fcurve.box,&sob->transform,&ob);
  nbshift=nb=ob;

  if(handle)
  {
   iconaddr(handle,sipmap[SIPX].realicon,&p);
   stringtounits(p,&nb.orgx,view->data.units);
  }
  if(index==SIPX) nb.orgx=sxsnap(view,nb.orgx+step);
  nbshift.orgx=(nb.orgx+=prefp(file)->rulerx);

  if(handle)
  {
   iconaddr(handle,sipmap[SIPY].realicon,&p);
   stringtounits(p,&nb.orgy,view->data.units);
  }
  if(index==SIPY) nb.orgy+=step;
  nbshift.orgy=(nb.orgy+=prefp(file)->rulery);

  if(handle)
  {
   iconaddr(handle,sipmap[SIPW].realicon,&p);
   stringtounits(p,&nb.width,view->data.units);
  }
  if(index==SIPW) nb.width+=step;

  if(handle)
  {
   iconaddr(handle,sipmap[SIPH].realicon,&p);
   stringtounits(p,&nb.height,view->data.units);
  }
  if(index==SIPH) nb.height+=step;

  obp=&ob;
  nbp=&nb;
  nbshiftp=&nbshift;

  err=transformselections(boxtrans,0,sipmodfn);
 }

 return(err);
}


static os_error * sipkey(int handle,int userhandle,int icon,int * key)
{
 os_error * err;
 viewstr  * view;
 int        ch;

 err=NULL;

 view=(viewstr*)userhandle;
 ch=*key;

 switch(ch)
 {
  case RETURN:
              err=sipaction(view,handle,0,-1);
              break;

      default:
              return(err);
 }

 *key=-1;

 return(err);
 USE(icon);
}





static int getselectionxns(int n)
{
 int          object;
 sobjectstr * sob;
 boxbitstr    bits;


 if(selectn && n<selectn && n>=0)
 {
  object=nselectedobject(n,selfile);
  sob=(sobjectstr*)fob_find(object,selfile);
  getboxbits(&sob->fcurve.box,&sob->transform,&bits);
  return(bits.orgx-prefp(selfile)->rulerx);
 }
 return(INT_MIN);
}


int getselectionxn(int fp)
{
 return(getselectionxns(stack[fp]));
}


int getselectionx(int fp)
{
 return(getselectionxns(0));
 USE(fp);
}




static int getselectionys(int n)
{
 int          object;
 sobjectstr * sob;
 boxbitstr    bits;


 if(selectn && n<selectn && n>=0)
 {
  object=nselectedobject(n,selfile);
  sob=(sobjectstr*)fob_find(object,selfile);
  getboxbits(&sob->fcurve.box,&sob->transform,&bits);
  return(bits.orgy-prefp(selfile)->rulery);
 }
 return(INT_MIN);
}



int getselectionyn(int fp)
{
 return(getselectionys(stack[fp]));
}


int getselectiony(int fp)
{
 return(getselectionys(0));
 USE(fp);
}




static int getselectionws(int n)
{
 int          object;
 sobjectstr * sob;
 boxbitstr    bits;

 if(selectn && n<selectn && n>=0)
 {
  object=nselectedobject(n,selfile);
  sob=(sobjectstr*)fob_find(object,selfile);
  getboxbits(&sob->fcurve.box,&sob->transform,&bits);
  return(bits.width);
 }
 return(INT_MIN);
}


int getselectionwn(int fp)
{
 return(getselectionws(stack[fp]));
}


int getselectionw(int fp)
{
 return(getselectionws(0));
 USE(fp);
}



static int getselectionhs(int n)
{
 int          object;
 sobjectstr * sob;
 boxbitstr    bits;

 if(selectn && n<selectn && n>=0)
 {
  object=nselectedobject(n,selfile);
  sob=(sobjectstr*)fob_find(object,selfile);
  getboxbits(&sob->fcurve.box,&sob->transform,&bits);
  return(bits.height);
 }
 return(INT_MIN);
}


int getselectionhn(int fp)
{
 return(getselectionhs(stack[fp]));
}


int getselectionh(int fp)
{
 return(getselectionhs(0));
 USE(fp);
}



static void setselectionareas(int n,int fp)
{
 int          object;
 sobjectstr * sob;
 boxbitstr    ob;
 boxbitstr    nb;
 boxbitstr    nbshift;

 if(selectn && n<selectn && n>=0)
 {
  object=nselectedobject(n,selfile);
  sob=(sobjectstr*)fob_find(object,selfile);

  getboxbits(&sob->fcurve.box,&sob->transform,&ob);
  nbshift=nb=ob;

  nb.orgx=stack[fp];
  nbshift.orgx=(nb.orgx+=prefp(selfile)->rulerx);

  nb.orgy=stack[fp+1];
  nbshift.orgy=(nb.orgy+=prefp(selfile)->rulery);

  nb.width=stack[fp+2];

  nb.height=stack[fp+3];

  obp=&ob;
  nbp=&nb;
  nbshiftp=&nbshift;

  zreport(transformselections(boxtrans,0,sipmodfn));
 }
}


void setselectionarean(int fp)
{
 setselectionareas(stack[fp],fp+1);
}


void setselectionarea(int fp)
{
 setselectionareas(0,fp);
 USE(fp);
}





static ipstr sipstr=
{
 NULL,           /* click    */
 sipkey,         /* key      */
 sipmap,         /* icon map */
 TSELECTTOOL,    /* template */
 NULL,           /* redraw   */
 siprefresh,     /* refresh  */
};



static os_error * sselect(toolboxstr * toolbox,viewstr * view,
                                               filestr * file,int select)
{
 ipchangepane(toolbox,IP_PANE1,select?&sipstr:NULL);

 return(NULL);
 USE(toolbox);
 USE(view);
 USE(file);
}



static os_error * sptr(toolboxstr * toolbox,posnstr * posn)
{
 if(findboxhandle(posn))  toolschangepointer(8,USERPOINTER,&resizeptr);
 else                     toolschangepointer(0,SYSTEMPOINTER,NULL);

 return(NULL);
 USE(toolbox);
}



static os_error * scancel(toolboxstr * toolbox,viewstr * view,filestr * file)
{
 aboxstr abox;

 if(magstate) return(magcancel(toolbox,view,file));


 if(selectstate || selectaltstate || selectddstate)
 {
  if(selectddstate)
  {
   abox.x0=scurrposn.xps-scurrposn.xpo;
   abox.x1=scurrposn.xpe-scurrposn.xpo;
   abox.y1=scurrposn.ype-scurrposn.ypo;
   abox.y0=scurrposn.yps-scurrposn.ypo;

   viewrefresharea(scurrposn.file,scurrposn.page,&abox,VRTOOLS|VRONELAYER);
  }
  else
  if(selectaltstate || selectstate!=STRANSLOCK)
  {
   abox.x0=sstartposn.xps-sstartposn.xpo;
   abox.x1=sstartposn.xpe-sstartposn.xpo;
   abox.y1=sstartposn.ype-sstartposn.ypo;
   abox.y0=sstartposn.yps-sstartposn.ypo;

   viewrefresharea(sstartposn.file,sstartposn.page,&abox,VRTOOLS|VRONELAYER);
  }

  srulermarks(0);

  selectstate=0;
  selectaltstate=0;
  selectddstate=0;
  toolsstopdrag();
  iprefreshpane(file,0x1,IP_PANE1);
 }

 return(NULL);

 USE(toolbox);
 USE(view);
 USE(file);
}


static os_error * smenu(toolboxstr * toolbox,viewstr * view,
                                             filestr * file,wimp_mousestr * m)
{
 return(openmenu("main_menu"));
 USE(view);
 USE(file);
 USE(m);
 USE(toolbox);
}


static int sxunitstep(viewstr * view)
{

 switch(view->data.units)
 {
  case  INCHES:
               return(72000/20);
               break;

  case  POINTS:
               return(3*1000);
               break;

  case  MILIMS:
       default:
               return(MMSIZE);
               break;

 }

}






static os_error * skey(toolboxstr * toolbox,int * key,fkeyfn * fk)
{
 os_error * err;
 int        ch;
 
 err=menukey(key);
 if(!err)
 {
  if(*key==-1) *fk=menukey;
  else
  {
   ch=*key;

   switch(ch)
   {
    case DEL:
             err=deleteselection2(menuview);
             break;

    case CLEFT:
    case ALT_CLEFT:
             err=sipaction(menuview,toolbox->iphandle[IP_PANE1],-sxunitstep(menuview),SIPX);
             break;

    case CRIGHT:
    case ALT_CRIGHT:
             err=sipaction(menuview,toolbox->iphandle[IP_PANE1],sxunitstep(menuview),SIPX);
             break;

    case CUP:
    case ALT_CUP:
             err=sipaction(menuview,toolbox->iphandle[IP_PANE1],-sxunitstep(menuview),SIPY);
             break;

    case CDOWN:
    case ALT_CDOWN:
             err=sipaction(menuview,toolbox->iphandle[IP_PANE1],sxunitstep(menuview),SIPY);
             break;

    case SHFT_CLEFT:
    case ALSH_CLEFT:
             err=sipaction(menuview,toolbox->iphandle[IP_PANE1],-10*sxunitstep(menuview),SIPX);
             break;

    case SHFT_CRIGHT:
    case ALSH_CRIGHT:
             err=sipaction(menuview,toolbox->iphandle[IP_PANE1],10*sxunitstep(menuview),SIPX);
             break;

    case SHFT_CUP:
    case ALSH_CUP:
             err=sipaction(menuview,toolbox->iphandle[IP_PANE1],-10*sxunitstep(menuview),SIPY);
             break;

    case SHFT_CDOWN:
    case ALSH_CDOWN:
             err=sipaction(menuview,toolbox->iphandle[IP_PANE1],10*sxunitstep(menuview),SIPY);
             break;


     default:
             return(err);
   }
   *key=-1;
   *fk=NULL;
  }
 }

 return(err);
 USE(toolbox);
}


static os_error * sselection(int eventn,int user,filestr * file,int object)
{
 if(toolcurrent(file)==TL_SELECT)
 {
  iprefreshpane(file,0x2,IP_PANE1);
 }

 return(NULL);
 USE(eventn);
 USE(user);
 USE(object);
}


static os_error * suselection(int eventn,int user,filestr * file,int object)
{
 if(toolcurrent(file)==TL_SELECT)
 {
  iprefreshpane(file,0x1,IP_PANE1);
 }

 return(NULL);
 USE(eventn);
 USE(user);
 USE(object);
}



static os_error * supref(int eventn,int user,viewstr * view,prefdatastr * old)
{
 filestr     * file;
 prefdatastr * xnew;

 file=view->file;

 if(toolcurrent(file)==TL_SELECT)
 {
  xnew=prefp(file);

  if(xnew->units!=old->units || xnew->rulerx!=old->rulerx || 
                                xnew->rulery!=old->rulery)
  {
   iprefreshpane(file,0x1,IP_PANE1);
  }
 }

 return(NULL);
 USE(eventn);
 USE(user);
}





static os_error * sinit(toolboxstr * toolbox,int code)
{
 if(code==TINITBOOT)
 {
  addevent(EVENT_OBJECTSELECTED1,(eventfn)sselection,0);
  addevent(EVENT_OBJECTDESELECTED1,(eventfn)sselection,0);
  addevent(EVENT_MODIFYSELECTION,(eventfn)suselection,0);
  addevent(EVENT_TRANSFORMSELECTION,(eventfn)suselection,0);
  addevent(EVENT_MOVESELECTION1,(eventfn)suselection,0);
  addevent(EVENT_TEXTFLOWSELECTION,(eventfn)suselection,0);
  addevent(EVENT_MASTERSELECTION,(eventfn)suselection,0);
  addevent(EVENT_PREFERENCES,(eventfn)supref,0);
 }
 else
 if(code==TINITCREATE) return(ipchangepane(toolbox,IP_PANE1,&sipstr));

 return(NULL);
}


static os_error * smess(toolboxstr * toolbox,filestr * file,
                                      int message,void * data)
{

 if(message==TSAVEPOSN) ssaveselection(file);

 return(NULL);

 USE(toolbox);
 USE(data);
}





static toolstr xselect=
{
 TL_SELECT,
 "select",
 "t_select\0t_pselect",
 "{SELTIP}",
 1,
 sclick,
 smove,
 sselect,
 sptr,
 scancel,
 sredraw,
 sdragend,
 smenu,
 skey,
 NULL,
 sinit,
 smess,
 0,

 1,       /*  preselect */
 0,       /*  noicon    */
 0,       /*  spare2    */
 0,       /*  spare3    */
 0,       /*  spare4    */ 
 0,       /*  spare5    */
 0,       /*  spare6    */
 0,       /*  spare7    */

 TPOINTER,
 0,
 0,
 0,       /* rank */
};



/*****************************************************************************/
/* Rotation tool */



typedef struct rptstr
{
 int ptr;
 int x;
 int y;
 int page;
} rptstr;


typedef struct rtoolstr
{
 rptstr rptr;
 int    centre;       /* show centre of rotation */

} rtoolstr;



static int          rdragstate;
static posnstr      rstartposn;
static transformstr rtransform;
static transformstr rinitial;
static int          rptrredraw;
static int          rptrskip;
static int          rptrdrag;
static rtoolstr   * rdragtool;
static rptstr       rold;
static rtoolstr     rtstart;




#define RIPCENTRE 3
#define RIPANGLE  0

static iconmap ripmap[]=
{
  51,0,0,IP_FSUB|IP_TIPS,  /* angle            0 */
  52,0,0,0,  /* dec angle        1 */
  53,0,0,0,  /* inc angle        2 */
  4,0,0,IP_TIPS,  /* crosshair on/off 3 */
  5,0,0,IP_TIPS,  /* matrix           4 */
  6,0,0,IP_TIPS,
  7,0,0,IP_TIPS,
  8,0,0,IP_TIPS,
  9,0,0,IP_TIPS,
 10,0,0,IP_TIPS,
 11,0,0,IP_TIPS,
 12,0,0,IP_TIPS,
 13,0,0,IP_TIPS,

 23,0,0,IP_TIPS,  /* 0                  */
 24,0,0,IP_TIPS,  /* 90                 */
 25,0,0,IP_TIPS,  /* 180                */
 26,0,0,IP_TIPS,  /* 270                */
 27,0,0,IP_TIPS,  /* 45                 */
 28,0,0,IP_TIPS,  /* 135                */
 29,0,0,IP_TIPS,  /* 225                */
 30,0,0,IP_TIPS,  /* 315                */

 0,0,0,0,      /* terminator */
};



int rotgetcentre(filestr * file,coordstr * cx)
{
 toolboxstr * toolbox;
 rtoolstr   * rtool;

 toolbox=file->toolbox;

 rtool=(rtoolstr*)toolbox->toolstate[toolindex(TL_ROTATE)].hprivate;

 cx->x=rtool->rptr.x;
 cx->y=rtool->rptr.y;

 return(rtool->centre);
}



/* want to rotate about centre */

static int rtrans(transformstr * r,transformstr * t,
                                    curvestr * c,transformstr * ct)
{
 coordstr     centre;
 transformstr temp1;

 /* find centre */

 if(rdragtool->rptr.ptr)
 {
  centre.x=rdragtool->rptr.x;
  centre.y=rdragtool->rptr.y;
 }
 else
 {
  centre.x=c->box.x+c->box.w/2;
  centre.y=c->box.y+c->box.h/2;
  tr_transformset(&centre,1,ct);     /* now got centre in real coords */
 }


 temp1=*t;
 temp1.e-=centre.x;
 temp1.f-=centre.y;

 tr_multiply(r,&temp1,&rtransform);

 r->e+=centre.x;
 r->f+=centre.y;

 return(TR_ROTATE);
}


#define ROTSIZE (2*MMSIZE)


static os_error * rrptr(viewredrawstr * v,int ptr,int x,int y)
{
 os_error * err;
 coordstr   c[2];
/* int        pal; */
 int        radius;
 int        dim;
 viewstr  * view;

 err=NULL;

 if(ptr)
 {
 /* pal=findcolourtint(v->file,&selcv); 
  rseteorcolpal(pal,WIMPDBLUE); */
  rseteorcol(WIMPWHITE,WIMPDBLUE);

  view=v->view;

  radius=scale(ROTSIZE,view->data.zoom.div,view->data.zoom.mul);
  dim=scale(3*MMSIZE,view->data.zoom.div,view->data.zoom.mul);

  c[0].x=c[1].x=x;
  c[0].y=y;
  c[1].y=y+radius;

  rcircle(c,0,v);

 /* rseteorcolpal(pal,WIMPBLACK); */
  rseteorcol(WIMPWHITE,WIMPBLACK);

  rbbc_move(x-dim,y,v);
  rbbc_draw(x+dim,y,v);
  rbbc_move(x,y-dim,v);
  rbbc_draw(x,y+dim,v);
 }

 return(err);
}




static os_error * rupdateptr(filestr * file,rtoolstr * rtool)
{
 os_error * err;

 err=NULL;

 if(rtool->rptr.ptr || rold.ptr)
 {
  rptrredraw=1;

  err=viewrefresharea(file,selpagek,NULL,VRTOOLS|VRONELAYER);

  rptrredraw=0;
 }

 rold.ptr=0;

 return(err);
}



static os_error * rredraw(toolboxstr * toolbox,viewredrawstr * v)
{
 rtoolstr * rtool;

 rtool=(rtoolstr*)toolbox->toolstate[toolindex(TL_ROTATE)].hprivate;

 if(rptrredraw && rold.page==v->page)
 {
  rrptr(v,rold.ptr,rold.x,rold.y);
 }
 else
 {
  if(rdragstate && rstartposn.page==v->page)
  {
   redrawselectiont(v,rtrans,0);
  }
 }

 if(!rptrskip && rtool->rptr.page==v->page) 
                   rrptr(v,rtool->rptr.ptr,rtool->rptr.x,rtool->rptr.y);

 return(NULL);
}



/* be careful not all object fields of posn may be set for groups */

static void rrotateobject(posnstr * posn,rtoolstr * rtool)
{
 sobjectstr * sob;
 coordstr     centre;

 rstartposn=*posn;  /* find centre of thing being rotated */

 sob=(sobjectstr*)fob_find(posn->object,posn->file);

 if(rtool->rptr.ptr)
 {
  centre.x=rtool->rptr.x;
  centre.y=rtool->rptr.y;
 }
 else
 {
  centre.x=sob->fcurve.box.x+sob->fcurve.box.w/2;
  centre.y=sob->fcurve.box.y+sob->fcurve.box.h/2;
  tr_transformset(&centre,1,&sob->transform);
                                         /* now got centre in real coords */
 }


 tr_setidentity(&rinitial);
 tr_setrotation(&rinitial,posn->x-centre.x,-(posn->y-centre.y));

 rstartposn.x=centre.x;
 rstartposn.y=centre.y;

 rdragstate=1;
 rdragtool=rtool;
 tr_setidentity(&rtransform);
 toolsstartdrag(posn,TOOLDRAGPAGE|TOOLDRAGKN|TOOLDRAGPTR);
}



static int findrotcentre(rtoolstr * rtool,posnstr * posn)
{
 viewstr * view;
 int       radius;

 view=posn->view;
 radius=scale(ROTSIZE,view->data.zoom.div,view->data.zoom.mul);

 if(rtool->rptr.ptr)
 {
  if((ABS(rtool->rptr.x-posn->x)<ROTSIZE) &&
     (ABS(rtool->rptr.y-posn->y)<ROTSIZE)) return(1);
 }
 return(0);
}



#define RTOPLEFT  0
#define RTOPMID   1
#define RTOPRIGHT 2
#define RLEFTMID  3
#define RMIDDLE   4
#define RRIGHTMID 5
#define RBOTLEFT  6
#define RBOTMID   7
#define RBOTRIGHT 8




static os_error * rposncross(rtoolstr * rtool,filestr * file,int posn)
{
 boxstr         box;
 transformstr   tr;
 sobjectstr   * sob;
 int            object;
 coordstr       cx;
 int            centre;
 int            master;
 objectbitstr   bits;
 int            claim;
 int            lock;
 int            nsel;


 rold=rtool->rptr;

 if((centre=rtool->centre)!=0)
 {
  nsel=nselected(&object,file);

//  cprintf("rposn nsel %d",nsel);

  if(nsel>0)
  {
   if(nsel==1)
   {
    sob=(sobjectstr*)fob_find(object,file);
    box=sob->fcurve.box;
    tr=sob->transform;
   }
   else 
   {
    getselectionbox(&box,file);
    tr_setidentity(&tr);
   }

   objectmaster(object,file,&master,&bits);
 
   claim=object;
   event(EVENT_ROTLOCK,&claim,(int)file);
   if(claim==object) 
      lock=((master && !bits.local) || bits.locked || bits.norotate);
   else
      lock=(claim==1);

   if(lock) centre=0;
  }
  else                 // case added to prevent access of non-existant objects
  {
   centre=0;
  }
 }

 if(centre)
 {
  rtool->rptr.page=selpagek;
  rtool->rptr.ptr=1;

  switch(posn)
  {
   case   RMIDDLE:
                  cx.x=box.x+box.w/2;
                  cx.y=box.y+box.h/2;
                  break;

   case  RTOPLEFT:
                  cx.x=box.x;
                  cx.y=box.y;
                  break;

   case  RBOTLEFT:
                  cx.x=box.x;
                  cx.y=box.y+box.h;
                  break;

   case RTOPRIGHT:
                  cx.x=box.x+box.w;
                  cx.y=box.y;
                  break;

   case   RTOPMID:
                  cx.x=box.x+box.w/2;
                  cx.y=box.y;
                  break;

   case   RBOTMID:
                  cx.x=box.x+box.w/2;
                  cx.y=box.y+box.h;
                  break;

   case  RLEFTMID:
                  cx.x=box.x;
                  cx.y=box.y+box.h/2;
                  break;

   case RRIGHTMID:
                  cx.x=box.x+box.w;
                  cx.y=box.y+box.h/2;
                  break;

   case RBOTRIGHT:
                  cx.x=box.x+box.w;
                  cx.y=box.y+box.h;
                  break;
  }

  tr_transformset(&cx,1,&tr);

  rtool->rptr.x=cx.x;
  rtool->rptr.y=cx.y;
 }
 else
 {
  rtool->rptr.ptr=0;
 }

 rupdateptr(file,rtool);

 return(NULL);
}




void rotupdate(filestr * file,int claim)
{
 rtoolstr   * rtool;
 toolboxstr * toolbox;
 int          index;

 toolbox=file->toolbox;
 index=toolindex(TL_ROTATE);

 if(toolbox->current==index)
 {
  rtool=(rtoolstr*)toolbox->toolstate[index].hprivate;

  if(claim) rposncross(rtool,file,RMIDDLE);

  iprefreshpane(file,0x1,IP_PANE1);
 }
}



static os_error * rselection(int eventn,int user,filestr * file,int object)
{
 rtoolstr   * rtool;

 toolboxstr * toolbox;
 int          index;

 index=toolindex(TL_ROTATE);
 toolbox=file->toolbox;
 if(toolbox->current==index)
 {
  rtool=(rtoolstr*)toolbox->toolstate[index].hprivate;
  rposncross(rtool,file,RMIDDLE);
  iprefreshpane(file,0x1,IP_PANE1);
 }

 return(NULL);
 USE(eventn);
 USE(user);
 USE(object);
}







static os_error * rclick(toolboxstr * toolbox,posnstr * posn)
{
 rtoolstr * rtool;
 int        object;
 int        fposn;
 int        claim;
 int        lock;


 if(!posn->ghost)
 {
  rtool=(rtoolstr*)toolbox->toolstate[toolindex(TL_ROTATE)].hprivate;
  rtstart=*rtool;

  fposn=findbox(posn,-1);

  if(posn->buttons==VBCLICKS || posn->buttons==VBCLICKA || posn->buttons==VBCLICKCS) // allow select + control 
  {
   if(!findrotcentre(rtool,posn)) sclicks(toolbox,posn,0);
  }
  else
  if(posn->buttons==VBDRAGS || posn->buttons==VBDRAGSS ||
     posn->buttons==VBLONGCLICKS)
  {
   if(findrotcentre(rtool,posn))
   {
    rptrdrag=1;
    toolsstartdrag(posn,TOOLDRAGPAGE|TOOLDRAGKN|TOOLDRAGPTR);
    rstartposn=*posn;
   }
   else
   if(fposn)
   {
    claim=posn->object;
    event(EVENT_ROTLOCK,&claim,(int)posn->file);
    if(claim==posn->object) lock=((posn->master && !posn->bits.local) || 
                                  posn->bits.locked || 
                                  posn->bits.norotate);
    else                    lock=(claim==1);

    if(!lock)
    {
     claim=0;

     event(EVENT_ROTSTART,&claim,(int)posn);

     if(!claim)
     {
      if((object=isselectedbig(posn->object,posn->file))!=0)
      {
       posn->object=object;
       rrotateobject(posn,rtool);

       rptrskip=1;
       viewrefreshall(posn->file,posn->view->branch,VRTOOLS|VRONELAYER);
       rptrskip=0;
      }
     }
    }
    else
    {
     toolschangepointer(128,USERPOINTER,&lockptr);
     toolsstartdrag(posn,TOOLDRAGWINDOW|TOOLDRAGPTR|TOOLDRAGNS);
    }
   }
  }
 }

 caretensure(posn->view);

 return(NULL);
 USE(toolbox);
}


static os_error * rmove(toolboxstr * toolbox,posnstr * posn,int move)
{
 aboxstr        abox;
 int            xshift;
 int            yshift;
 transformstr   rcurrent;
 rtoolstr     * rtool;

 rtool=(rtoolstr*)toolbox->toolstate[toolindex(TL_ROTATE)].hprivate;

 toolschangepointer(0,USERPOINTER,&rotateptr);

 if(move && posn->page==rstartposn.page && !posn->ghost)
 {
  if(rdragstate)
  {
   abox.x0=rstartposn.xps-rstartposn.xpo;
   abox.x1=rstartposn.xpe-rstartposn.xpo;
   abox.y1=rstartposn.ype-rstartposn.ypo;
   abox.y0=rstartposn.yps-rstartposn.ypo;

   rptrskip=1;
   viewrefresharea(posn->file,posn->page,&abox,VRTOOLS|VRONELAYER);
  
   xshift=posn->x-rstartposn.x;
   yshift=posn->y-rstartposn.y;

   tr_setidentity(&rcurrent);
   tr_setrotation(&rcurrent,xshift,yshift);
   tr_multiply(&rtransform,&rinitial,&rcurrent);

   viewrefresharea(posn->file,posn->page,&abox,VRTOOLS|VRONELAYER);
   rptrskip=0;
   iprefreshpane(posn->file,0x1,IP_PANE1);
  }
  else
  if(rptrdrag)
  {
   rold=rtool->rptr;

   xshift=posn->x-rstartposn.x;
   yshift=posn->y-rstartposn.y;

   if(rstartposn.shift)
   {
    if((ABS(xshift))>(ABS(yshift))) yshift=0;
    else                            xshift=0;
   }

   rtool->rptr.x=rstartposn.x+xshift;
   rtool->rptr.y=rstartposn.y+yshift;

   rtool->rptr.x=gridsnapx(rtool->rptr.x,posn->view,0);
   rtool->rptr.y=gridsnapy(rtool->rptr.y,posn->view,0);
   rupdateptr(posn->file,rtool);
  }
 }

 return(NULL);
}



static os_error * rdragend(toolboxstr * toolbox,posnstr * posn,wimp_box * box)
{
 os_error * err;
 aboxstr    abox;

 err=NULL;

 if(rdragstate)
 {
  abox.x0=rstartposn.xps-rstartposn.xpo;
  abox.x1=rstartposn.xpe-rstartposn.xpo;
  abox.y1=rstartposn.ype-rstartposn.ypo;
  abox.y0=rstartposn.yps-rstartposn.ypo;

  rptrskip=1;
  viewrefresharea(rstartposn.file,rstartposn.page,&abox,VRTOOLS|VRONELAYER);
  rdragstate=0;
  rptrskip=0;

  err=transformselection(rtrans,0);
  iprefreshpane(posn->file,0x1,IP_PANE1);
 }
 else
 if(rptrdrag)
 {
  rptrdrag=0;
 }


 return(err);

 USE(posn);
 USE(box);
 USE(toolbox);
}





static os_error * rkey(toolboxstr * toolbox,int * key,fkeyfn * fk)
{
 os_error * err;
 int        ch;
 
 err=menukey(key);
 if(!err)
 {
  if(*key==-1) *fk=menukey;
  else
  {
   ch=*key;

   if(ch==DEL)
   {
    err=deleteselection2(menuview);
    *key=-1;
    *fk=NULL;
   }
  }
 }


 return(err);
 USE(toolbox);
}




static os_error * rsetangle(rtoolstr * rtool,viewstr * view,
                                             int handle,int angle)
{
 os_error        * err;
 int               object;
 sobjectstr      * sob;
 transformbitstr   bits;
 int               data[2];
 char              string[12];

 err=NULL;

 data[0]=0;
 data[1]=angle;
 err=event(EVENT_ROTDO,view->file,(int)data);
 if(!data[0] && nselected(&object,view->file))
 {
  sob=(sobjectstr*)fob_find(object,view->file);

  tr_getbits(&sob->transform,&bits);

  rdragtool=rtool;

  tr_setidentity(&rtransform);
  tr_setrotationa(&rtransform,angle-bits.angle);

  err=transformselection(rtrans,0);

  writeangle(angle,string);
  writeiconc(handle,ripmap[RIPANGLE].realicon,string);
 }

 return(err);
}




static os_error * rchangeangle(rtoolstr * rtool,viewstr * view,
                                                    int handle,int step)
{
 os_error        * err;
 char            * p;
 int               angle;

 err=iconaddr(handle,ripmap[RIPANGLE].realicon,&p);
 if(!err)
 {
  err=readangle(&angle,p);
  if(!err)
  {
   if(step) angle=snapangle(angle+step*0x10000,step*0x10000);
   err=rsetangle(rtool,view,handle,angle);
  }
 }

 return(err);
}






static os_error * ripclick(int handle,int userhandle,wimp_mousestr * m)
{
 os_error   * err;
 viewstr    * view;
 filestr    * file;
 toolboxstr * toolbox;
 rtoolstr   * rtool;
 int          step;

 err=NULL;
 view=(viewstr *)userhandle;
 file=view->file;
 toolbox=file->toolbox;
 rtool=(rtoolstr*)toolbox->toolstate[toolindex(TL_ROTATE)].hprivate;

 switch(m->i)
 {
  case  52: /* dec angle */
  case  53: /* inc angle */

          if(m->i==52) step=-1;
          else         step=1;

          if(m->bbits==0x1) step=-step;
          if(isshift) step*=10;
          err=rchangeangle(rtool,view,handle,step);
          break;

  case  4:       /* cross hair on/off */
          selectst(handle,ripmap[RIPCENTRE].realicon,rtool->centre^=1);
          err=rposncross(rtool,file,RMIDDLE);
          break;

  default:
          if(m->i>=23 && m->i<=26)
              err=rsetangle(rtool,view,handle,-90*0x10000*(m->i-23));
          else
          if(m->i>=27 && m->i<=30)
              err=rsetangle(rtool,view,handle,-(45+90*(m->i-27))*0x10000);
          else
          if(m->i>=5 && m->i<=13) 
              err=rposncross(rtool,file,m->i-5+RTOPLEFT);
          break;

 }

 return(err);
}




static os_error * ripkey(int handle,int userhandle,int icon,int * key)
{
 os_error * err;
 viewstr    * view;
 filestr    * file;
 toolboxstr * toolbox;
 rtoolstr   * rtool;
 int        ch;
 wimp_mousestr m;

 err=NULL;

 ch=*key;

 switch(ch)
 {
     case CUP:
   case CDOWN:
              if(icon==51)
              {
               m.i=(ch==CUP)?53:52;
               m.w=handle;
               m.bbits=wimp_BLEFT;
               ripclick(handle,userhandle,&m);
              }
              else return(err);

              break;
 
  case RETURN:
              view=(viewstr *)userhandle;
              file=view->file;
              toolbox=file->toolbox;
              rtool=(rtoolstr*)toolbox->toolstate[toolindex(TL_ROTATE)].hprivate;
              rchangeangle(rtool,view,handle,0);
              break;

      default:
              return(err);
 }

 *key=-1;

 return(err);
}





static os_error * riprefresh(viewstr * view,int handle,int code)
{
 int                object;
 sobjectstr       * sob;
 transformbitstr    bits;
 transformstr       tr;
 filestr          * file;
 toolboxstr       * toolbox;
 rtoolstr         * rtool;
 int                data[2];
 char               string[12];

 file=view->file;
 toolbox=file->toolbox;
 rtool=(rtoolstr*)toolbox->toolstate[toolindex(TL_ROTATE)].hprivate;

 if(code==-1) 
 {
  writeiconf(view->info,IP_HELPI,"%s","{ROT00}");
  selectst(handle,ripmap[RIPCENTRE].realicon,rtool->centre);
 }

 if(code & 0x1)
 {
  data[0]=0;
  event(EVENT_ROTANGLE,file,(int)data);
  if(data[0])
  {
   writeangle(data[1],string);
   writeiconc(handle,ripmap[RIPANGLE].realicon,string);
  }
  else
  if(nselected(&object,view->file))
  {
   sob=(sobjectstr*)fob_find(object,view->file);

   if(rdragstate)
   {
    tr_multiply(&tr,&sob->transform,&rtransform);
    tr_getbits(&tr,&bits);
   }
   else
   {
    tr_getbits(&sob->transform,&bits);
   }

   writeangle(bits.angle,string);
   writeiconc(handle,ripmap[RIPANGLE].realicon,string);
  }
  else writeicon(handle,ripmap[RIPANGLE].realicon,"");
 }

 return(NULL);
}


static ipstr ripstr=
{
 ripclick,       /* click    */
 ripkey,         /* key      */
 ripmap,         /* icon map */
 TROTATETOOL,    /* template */
 NULL,           /* redraw   */
 riprefresh,     /* refresh  */

};



static os_error * rselect(toolboxstr * toolbox,viewstr * view,
                                               filestr * file,int select)
{
 rtoolstr * rtool;

 rtool=(rtoolstr*)toolbox->toolstate[toolindex(TL_ROTATE)].hprivate;

 if(select)
 {
  rtool->rptr.ptr=0;
  rposncross(rtool,file,RMIDDLE);
 }
 else
 {
  rupdateptr(file,rtool);
  rtool->rptr.ptr=0;
 }

 ipchangepane(toolbox,IP_PANE1,select?&ripstr:NULL);

 return(NULL);

 USE(view);
}



static os_error * rptr(toolboxstr * toolbox,posnstr * posn)
{
 toolschangepointer(0,USERPOINTER,&rotateptr);

 return(NULL);

 USE(posn);
 USE(toolbox);
}



static os_error * rcancel(toolboxstr * toolbox,viewstr * view,filestr * file)
{
 aboxstr    abox;
 rtoolstr * rtool;

 if(rdragstate)
 {
  abox.x0=rstartposn.xps-rstartposn.xpo;
  abox.x1=rstartposn.xpe-rstartposn.xpo;
  abox.y1=rstartposn.ype-rstartposn.ypo;
  abox.y0=rstartposn.yps-rstartposn.ypo;

  rptrskip=1;
  viewrefresharea(rstartposn.file,rstartposn.page,&abox,VRTOOLS|VRONELAYER);
  rdragstate=0;
  toolsstopdrag();
  rptrskip=0;

  iprefreshpane(file,0x1,IP_PANE1);
 }
 else
 if(rptrdrag)
 {
  rtool=(rtoolstr*)toolbox->toolstate[toolindex(TL_ROTATE)].hprivate;
  rold=rtool->rptr;
  *rtool=rtstart;

  rupdateptr(file,rtool);

  rptrdrag=0;
 }

 return(NULL);

 USE(view);
 USE(toolbox);
}





static os_error * ruchange(int eventn,int user,filestr * file,int object)
{
 if(toolcurrent(file)==TL_ROTATE)
 {
  iprefreshpane(file,0x1,IP_PANE1);
 }

 return(NULL);
 USE(eventn);
 USE(user);
 USE(object);
}




static os_error * rinit(toolboxstr * toolbox,int code)
{
 rtoolstr * rtool;

 if(code==TINITBOOT)
 {
  addevent(EVENT_OBJECTSELECTED1,(eventfn)rselection,0);
  addevent(EVENT_OBJECTDESELECTED1,(eventfn)rselection,0);
  addevent(EVENT_TRANSFORMSELECTION,(eventfn)ruchange,0);
  addevent(EVENT_ROTCHANGE,(eventfn)ruchange,0);
 }
 else
 if(code==TINITCREATE)
 {
  rtool=(rtoolstr*)toolbox->toolstate[toolindex(TL_ROTATE)].hprivate;
  rtool->centre=1;
 }

 return(NULL);
}





static toolstr rotate=
{
 TL_ROTATE,
 "rotate",
 "t_rotate\0t_protate",
 "{ROTTIP}",
 1,
 rclick,
 rmove,
 rselect,
 rptr,
 rcancel,
 rredraw,
 rdragend,
 smenu,
 rkey,
 NULL,
 rinit,
 smess,
 0,

 0,       /*  preselect */
 0,       /*  noicon    */
 0,       /*  spare2    */
 0,       /*  spare3    */
 0,       /*  spare4    */ 
 0,       /*  spare5    */
 0,       /*  spare6    */
 0,       /*  spare7    */

 TPOINTER,
 0,
 sizeof(rtoolstr),
 4,       /* rank       */
};




/*****************************************************************************/
/* magnify tool */


/*

#define MAGSIZE 8

typedef struct magstackstr
{
 viewstr    * view;
 magzoomstr   zoom[MAGSIZE];
 int          zoomopt[MAGSIZE];
 int          first;
 int          last;

} magstackstr;


static magstackstr * magstack;
static int           magsize;

*/



#define MIPZOOM  0

static iconmap mipmap[]=
{
 51,0,0,IP_FSUB|IP_TIPS,  /* zoom            0 */
 52,0,0,IP_TIPS,  /* dec zoom        1 */
 53,0,0,IP_TIPS,  /* inc zoom        2 */
 6,0,0,IP_TIPS,  /* prev zoom       3 */
 8,0,0,IP_TIPS,  /* 50%             4 */
 13,0,0,IP_TIPS, /* Fit screen      5 */
 14,0,0,IP_TIPS, /* Fit window      6 */
 15,0,0,IP_TIPS, /* 100%            7 */
 16,0,0,IP_TIPS, /* 200%            8 */
 18,0,0,IP_TIPS, /* 75%             9 */
 20,0,0,IP_TIPS, /* 150%           10 */
 22,0,0,IP_TIPS, /* 400%           11 */
 25,0,0,IP_TIPS, /* Fit width      12 */
 26,0,0,IP_TIPS, /* Fit height     13 */


 0,0,0,0,  /* terminator */
};


/* works on actual real current zoom */

static os_error * mstepzoom(posnstr * posn,int step)
{
 int       z;
 viewstr * view;

 view=posn->view;

 z=(view->data.zoom.mul*100)/view->data.zoom.div;
 /*z+=step; */
 if(step>0) z*=2;
 else       z/=2;

 z=CUT(z,ZOOMMIN,ZOOMMAX);

 view->data.defzoom.mul=z;
 view->data.defzoom.div=100;

 return(zoomdef(view,posn->mousex-view->mainwindow.x0,
                     posn->mousey-view->mainwindow.y1));//mainpane
}


/* works on value read from info pal icon */

static os_error * mchangezoom(viewstr * view,int handle,int step)
{
 os_error * err;
 int        z;
 char     * p;

 err=iconaddr(handle,mipmap[MIPZOOM].realicon,&p);
 if(!err)
 {
  err=stringtonumber(p,&z);
  if(!err)
  {
   z+=step;
   z=CUT(z,ZOOMMIN,ZOOMMAX);

   view->data.defzoom.mul=z;
   view->data.defzoom.div=100;

   err=zoomdef(view,0,0);
  }
 }

 return(err);
}










static os_error * magptr(toolboxstr * toolbox,posnstr * posn)
{
 toolschangepointer(0,USERPOINTER,&magnifyptr);

 return(NULL);

 USE(toolbox);
 USE(posn);
}



static os_error * magdragend(toolboxstr * toolbox,posnstr * posn,
                                                       wimp_box * box)
{
 os_error * err;
 aboxstr    abox;
 viewstr  * view;
 int        w;
 int        h;
 int        ww=0;
 int        wh=0;
 int        z;

 err=NULL;


 if(magstate)
 {
  abox.x0=sstartposn.xps-sstartposn.xpo;
  abox.x1=sstartposn.xpe-sstartposn.xpo;
  abox.y1=sstartposn.ype-sstartposn.ypo;
  abox.y0=sstartposn.yps-sstartposn.ypo;

  viewrefresharea(posn->file,posn->page,&abox,VRTOOLS);

  magstate=0;

  view=posn->view;

  w=(scurrposn.x-sstartposn.x);
  h=(sstartposn.y-scurrposn.y);
  w=ABS(w);
  h=ABS(h);
  w=osunits(w,view);
  h=osunits(h,view);


  if(w>(2*cvdu.deltax) && h>(2*cvdu.deltay))
  {
   ww=(view->mainwindow.x1-view->mainwindow.x0)-(view->left+view->right); //mainpane
   wh=(view->mainwindow.y1-view->mainwindow.y0)-(view->top+view->bottom);

   z=(view->data.zoom.mul*100)/view->data.zoom.div;
  
   if(w*wh<h*ww)  /* w/ww < h/wh */
   {
    z=(z*wh)/h;
   }
   else
   {
    z=(z*ww)/w;
   }
  }
  else
  {
   z=0;
  }

  if(z==0 || z>=2*ZOOMMAX)
  {
   if(sstartposn.buttons==VBCLICKS) mstepzoom(&sstartposn,50);                     
   else
   if(sstartposn.buttons==VBCLICKA) mstepzoom(&sstartposn,-50);
  }
  else
  {
   z=CUT(z,ZOOMMIN,ZOOMMAX);

   view->data.defzoom.mul=z;
   view->data.defzoom.div=100;

   err=zoomdef2(view,
                     (view->mainwindow.x1-ww/2)-view->mainwindow.x0,
                     (view->mainwindow.y0+wh/2)-view->mainwindow.y1,
                scurrposn.mousex-osunits(scurrposn.x-sstartposn.x,view)/2
                                       -(view->mainwindow.x1-ww/2),
                scurrposn.mousey+osunits(scurrposn.y-sstartposn.y,view)/2
                                       -(view->mainwindow.y0+wh/2)); //mainpane
  }
 }

 return(err);

 USE(toolbox);
 USE(box);
}



static os_error * magcancel(toolboxstr * toolbox,viewstr * view,filestr * file)
{
 aboxstr abox;

 USE(view);
 USE(file);

 if(magstate)
 {
  abox.x0=sstartposn.xps-sstartposn.xpo;
  abox.x1=sstartposn.xpe-sstartposn.xpo;
  abox.y1=sstartposn.ype-sstartposn.ypo;
  abox.y0=sstartposn.yps-sstartposn.ypo;

  viewrefresharea(sstartposn.file,sstartposn.page,&abox,VRTOOLS);
  magstate=0;
  toolsstopdrag();
 }

 return(NULL);
 USE(toolbox);
}



static os_error * magmove(toolboxstr * toolbox,posnstr * posn,int move)
{
 aboxstr      abox;

 if(move)
 {
  if(magstate)
  {
   if(posn->page==sstartposn.page && !posn->ghost)
   {
    abox.x0=sstartposn.xps-sstartposn.xpo;
    abox.x1=sstartposn.xpe-sstartposn.xpo;
    abox.y1=sstartposn.ype-sstartposn.ypo;
    abox.y0=sstartposn.yps-sstartposn.ypo;

    viewrefresharea(posn->file,posn->page,&abox,VRTOOLS);
    scurrposn=*posn;
    viewrefresharea(posn->file,posn->page,&abox,VRTOOLS);
    iprefreshpane(posn->file,0x1,IP_PANE1);

    toolschangepointer(0,USERPOINTER,&magnifyptr);
   }
  }
 }

 return(NULL);
 USE(toolbox);
}


static os_error * magredraw(toolboxstr * toolbox,viewredrawstr * v)
{
 if(magstate && sstartposn.page==v->page)
 {
  rseteorcol(WIMPWHITE,WIMPGREY2);
  bbc_rectangle(
           v->ox+osunits(sstartposn.xpo+sstartposn.x,v->view),
           v->oy-osunits(sstartposn.ypo+sstartposn.y,v->view),
               osunits(scurrposn.x-sstartposn.x,v->view),
               osunits(sstartposn.y-scurrposn.y,v->view));
 }

 return(NULL);
 USE(toolbox);
}



static os_error * mkey(toolboxstr * toolbox,int * key,fkeyfn * fk)
{
 menukey(key);
 if(*key==-1) *fk=menukey;

 return(NULL);
 USE(toolbox);
}



static os_error * magclick(toolboxstr * toolbox,posnstr * posn)
{
 int buttons;


 if(!posn->ghost)
 {
  buttons=posn->buttons;
  buttons&=~(VBALT|VBCTRL);

  if(buttons==VBCLICKS || buttons==VBCLICKA)
  {
   sstartposn=*posn;
   scurrposn=*posn;
   sstartposn.buttons=buttons;
   toolsstartdrag(posn,TOOLDRAGPAGE|TOOLDRAGKN|TOOLDRAGPTR);
   magstate=1;
  }
  else
  if(buttons==VBDBLCLICKS)
  {
   mstepzoom(posn,50);
  }
  else
  if(buttons==VBDBLCLICKA)
  {
   mstepzoom(posn,-50);
  }
 }

 caretensure(posn->view);

 return(NULL);
 USE(toolbox);
}




static os_error * mipclick(int handle,int userhandle,wimp_mousestr * m)
{
 os_error   * err;
 viewstr    * view;
 int          step;

 err=NULL;
 view=(viewstr *)userhandle;

 switch(m->i)
 {
  case  52: /* dec zoom */
  case  53: /* inc zoom */

          if(m->i==52) step=-1;
          else         step=1;

          if(m->bbits==0x1) step=-step;
          if(isshift) step*=10;
          err=mchangezoom(view,handle,step);
          break;

  case  6:
          view->data.defzoom=view->data.prevzoom;
          err=zoomopt(view,view->data.prevzoomopt);
          break;

  case  8:
          err=zoomopt(view,ZOOM50);
          break;

  case 15:
          err=zoomopt(view,ZOOM100);
          break;

  case 16:
          err=zoomopt(view,ZOOM200);
          break;

  case 14:
          err=zoomopt(view,ZOOMFIT);
          break;

  case 13:
          err=zoomopt(view,ZOOMSCR);
          break;

  case 18:
          err=zoomopt(view,ZOOM75);
          break;

  case 20:
          err=zoomopt(view,ZOOM150);
          break;

  case 22:
          err=zoomopt(view,ZOOM400);
          break;

  case 25:
          err=zoomopt(view,ZOOMSCRW);
          break;

  case 26:
          err=zoomopt(view,ZOOMSCRH);
          break;
 }

 return(NULL);
}



static os_error * mipkey(int handle,int userhandle,int icon,int * key)
{
 os_error * err;
 viewstr  * view;
 int        ch;
 wimp_mousestr m;

 err=NULL;


 ch=*key;

 switch(ch)
 {
     case CUP:
   case CDOWN:
              if(icon==51)
              {
               m.i=(ch==CUP)?53:52;
               m.w=handle;
               m.bbits=wimp_BLEFT;
               mipclick(handle,userhandle,&m);
              }
              else return(err);

              break;
 
  case RETURN:
              view=(viewstr *)userhandle;
              mchangezoom(view,handle,0);
              break;

      default:
              return(err);
 }

 *key=-1;

 return(err);
}





static os_error * miprefresh(viewstr * view,int handle,int code)
{
 filestr    * file;

 file=view->file;

 if(code==-1)
 {
  writeiconf(view->info,IP_HELPI,"%s","{MAG00}");
 }

 if(code)
 {
  writeiconfc(handle,mipmap[MIPZOOM].realicon,"%d",
                              (view->data.zoom.mul*100)/view->data.zoom.div);
 }

 return(NULL);
}




static ipstr mipstr=
{
 mipclick,       /* click    */
 mipkey,         /* key      */
 mipmap,         /* icon map */
 TMAGNIFYTOOL,   /* template */
 NULL,           /* redraw   */
 miprefresh,     /* refresh  */
};





static os_error * magselect(toolboxstr * toolbox,viewstr * view,
                                           filestr * file,int select)
{
 if(select)
 {
  view=file->view;
  while(view)
  {
   view->data.prevzoom=view->data.defzoom;
   view->data.prevzoomopt=view->data.zoomopt;
   view=view->next;
  }
 }

 ipchangepane(toolbox,IP_PANE1,select?&mipstr:NULL);

 return(NULL);
}


static os_error * mzoomevent(int eventn,int user,filestr * file,viewstr * view)
{
 if(toolcurrent(file)==TL_MAGNIFY && view->info)
 {
  iprefreshpane(file,0x1,IP_PANE1);
 }

 return(NULL);

 USE(eventn);
 USE(user);
}



static os_error * minit(toolboxstr * toolbox,int code)
{
 if(code==TINITBOOT)
 {
  addevent(EVENT_ZOOM,(eventfn)mzoomevent,0);
 }

 return(NULL);
 USE(toolbox);
}





static toolstr magnify=
{
 TL_MAGNIFY,
 "magnify",
 "t_magnify\0t_pmagnify",
 "{MAGTIP}",
 1,
 magclick,
 magmove,
 magselect,
 magptr,
 magcancel,
 magredraw,
 magdragend,
 smenu,
 mkey,
 NULL,
 minit,
 smess,
 0,

 0,       /*  preselect */
 0,       /*  noicon    */
 0,       /*  spare2    */
 0,       /*  spare3    */
 0,       /*  spare4    */ 
 0,       /*  spare5    */
 0,       /*  spare6    */
 0,       /*  spare7    */


 TPOINTER,
 0,
 0,
 7,       /*  rank      */
};




/*****************************************************************************/



os_error * toolboxinit(void)
{
 os_error * err;

 err=toolsinit();

 if(!err) err=toolregister(&xselect);
 if(!err) err=toolregister(&rotate);
 if(!err) err=toolregister(&magnify);
 if(!err) err=frametoolinit();
 if(!err) err=patheditinit();

 if(!err) err=infopalinit();

 return(err);
}

