/*->c.impex */


#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "bbc.h"
#include "transform.h"
#include "xprocess.h"

#include "err.h"
#include "filetype.h"
#include "fsx.h"
#include "ram.h"
#include "bf.h"
#include "xext.h"
#include "poll.h"
#include "alloc.h"
#include "trans.h"
#include "colour.h"
#include "etc.h"
#include "bits.h"
#include "save.h"
#include "scrap.h"
#include "deb.h"
#include "oscbx.h"


#include "constants.h"
#include "str.h"

#include "reslink.h"
#include "obj.h"
#include "object.h"
#include "dscript.h"
#include "posn.h"
#include "file.h"
#include "view.h"
#include "dbfile.h"
#include "transload.h"
#include "clip.h"


#include "pxstr.h"
#include "pict.h"
#include "pxref.h"


#include "impex.h"



static impexstr * impexs;
static int        impexn;


#define TCHUNK 16*sizeof(impexstr)


os_error * impexregister(impexstr * impex)
{
 os_error * err;

 if(impexs) 
    err=flex_chunk((flex_ptr)&impexs,(impexn+1)*sizeof(impexstr),TCHUNK);
 else
    err=flex_alloc((flex_ptr)&impexs,TCHUNK);

 if(!err)
 {
  impexs[impexn]=*impex;
  impexn++;
 }

 return(err);
}

static impexstr * findimpexstr(int ftype)
{
 int i;

 ftype&=IMPFTYPE;

 for(i=0;i<impexn;i++) if(impexs[i].ftype==ftype) return(&impexs[i]);
 return(NULL);
}


static impexstr * findimpexname(char * name)
{
 int i;

 for(i=0;i<impexn;i++) 
 {
  if(!cstrcmp(impexs[i].xtypename,name)) return(&impexs[i]);
 }

 return(NULL);
}



os_error * importloadpost(int ftype,int object,filestr * file)
{
 os_error * err;
 impexstr * impex;
 impex=findimpexstr(ftype);
 if(impex && impex->importpost) err=impex->importpost(ftype,object,file);
 else                           err=geterror(ENSFL);
 return(err);
}


os_error * importload(char * name,int ftype,int xv,int * object,filestr * file)
{
 os_error * err;
 impexstr * impex;

	cprintf("importload %s",name);

 err=NULL;
 *object=0;

 impex=findimpexstr(ftype);
 if(impex)
 {
  if(impex && impex->import) err=impex->import(name,ftype,xv,object,file);
  else                       err=geterror(ENSFL);
 }
 else
 if(xv==0)
 {
  int        transhandle;
  char       transname[FILENAMELEN];
  void     * eh;


  cprintf("importload trans view");

  err=sc_create(&transhandle);
  if(!err)
  {
   strcpy(transname,sc_name(transhandle));
   err=transload(name,transname,&ftype);
   if(!err)
   {
    impex=findimpexstr(ftype);

    if(impex && impex->import) err=impex->import(transname,ftype,xv,object,file);
    else                       err=geterror(ENSFL);
   }
 
   eh=errsave(err);
   fs_delete(sc_name(transhandle));
   sc_remove(transhandle,err);
   err=errload(eh,NULL);
  }
 }
 else
 {
  err=geterror(ENSFL);
 }

 return(err);
}



os_error * importreferenceload(char * name,int ftype,int * object,filestr * file)
{
 os_error * err;
 impexstr * impex;
 char       expandedname[FILENAMELEN];
	int        i;

 err=pxexpandpath(expandedname,NITEM(expandedname),name,file);
 if(!err)
 {
  impex=findimpexstr(ftype);
  if(impex)
  {
   if(impex->reference) err=impex->reference(name,expandedname,ftype,object,file);
   else                 err=geterror(ENSFL);
  }
  else
  {
 	 for(i=0;i<impexn;i++)
   {
				impex=impexs+i;
    if(impex->data & IMPXPICT)
				{
					ftype=impexs[i].ftype;
   		err=translatetest(name,ftype);
     if(!err) break;
				}
			}

   if(i<impexn && impex->reference) err=impex->reference(name,expandedname,ftype,object,file);
   else                             err=geterror(ENSFL);
  }
 }

 return(err);
}



os_error * importreferencereload(char * name,char * expname,int ftype,
                                             int object,filestr * file)
{
 os_error * err;
 impexstr * impex;
	int        i;

 err=NULL;


 impex=findimpexstr(ftype);
 if(impex)
 {
  if(impex->referencereload) err=impex->referencereload(name,expname,ftype,object,file);
  else                       err=geterror(ENSFL);
 }
 else
 {
	 for(i=0;i<impexn;i++)
  {
			impex=impexs+i;
   if(impex->data & IMPXPICT)
			{
				ftype=impexs[i].ftype;
  		err=translatetest(name,ftype);
    if(!err) break;
			}
		}

  if(i<impexn && impex->reference) err=impex->referencereload(name,expname,ftype,object,file);
  else                             err=geterror(ENSFL);
 }

 return(err);
}



os_error *importmakeproxy(int ftype,int object,filestr *file,pxproxystr *proxy)
{
 os_error * err;
 impexstr * impex;
 impex=findimpexstr(ftype);
 if(impex && impex->makeproxy) err=impex->makeproxy(ftype,object,file,proxy);
 else                          err=geterror(ENSFL);
 return(err);
}



os_error * importremoveproxy(int ftype,int object,filestr * file)
{
 os_error * err;
 impexstr * impex;
 impex=findimpexstr(ftype);
 if(impex && impex->removeproxy)
                               err=impex->removeproxy(ftype,object,file);
 else                          err=geterror(ENSFL);
 return(err);
}




os_error * importembed(int ftype,int object,filestr * file)
{
 os_error * err;
 impexstr * impex;
 impex=findimpexstr(ftype);
 if(impex && impex->importembed)
                               err=impex->importembed(ftype,object,file);
 else                          err=geterror(ENSFL);
 return(err);
}




os_error * render(aboxstr * abox,transformstr * tr,
                  processstr * pr,picturestr * picturep,viewredrawstr * v)
{
 os_error     * err;
 impexstr     * impex;
 pxproxyimstr * proxy;
 int            olduseproxy;

 olduseproxy=pr->useproxies;

 if(pr->useproxies && picturep->proxylen)
 {
  proxy=(pxproxyimstr*)(picturep->data+((picturep->imagelen+0x3)&(~0x3)));
  impex=findimpexstr(proxy->ftype);
  if(!impex || !impex->render) // if we can't render the proxy
  {                            // have a go at rendering the image
   pr->useproxies=0;
   impex=findimpexstr(picturep->ftype);
  }
 }
 else
 {
  impex=findimpexstr(picturep->ftype);
  if(!impex || !impex->render)  // if we can't render the image
  {                             // render the proxy if there is one
   if(picturep->proxylen)
   {
    pr->useproxies=1;
    proxy=(pxproxyimstr*)(picturep->data+((picturep->imagelen+0x3)&(~0x3)));
    impex=findimpexstr(proxy->ftype);
   }
  }
 }

// cprintf("type %x impex %x render %x",picturep->ftype,impex,impex?impex->render:0);

 if(impex && impex->render) err=impex->render(abox,tr,pr,picturep,v);
 else                       err=NULL;

 pr->useproxies=olduseproxy;
 
 return(err);
}



os_error * imcache(int picture,filestr * file)
{
 os_error   * err;
 picturestr * picturep;
 impexstr   * impex;

	err=NULL;

	if(picture)
	{
  picturep=(picturestr*)fob_find(picture,file);
  impex=findimpexstr(picturep->ftype);
  if(impex && impex->cache)  err=impex->cache(picturep,file);
  else                       err=NULL;
	}

 return(err);
}



os_error * imlist(int picture,filestr * file)
{
 os_error   * err;
 picturestr * picturep;
 impexstr   * impex;

// return(NULL); //dpfix

 impex=NULL;
 if(picture)
 {
  picturep=(picturestr*)fob_find(picture,file);
  impex=findimpexstr(picturep->ftype);
 }

// cprintf("imlist impex %x ftype %x listfonts %x",impex,picturep->ftype,impex->listfonts);

 if(impex && impex->listfonts)  err=impex->listfonts(picturep,file);
 else                           err=NULL;

 return(err);
}






int imrendermem(picturestr * picturep,
                     transformstr * tr,processstr * pr,viewredrawstr * v)
{
 impexstr   * impex;

 impex=findimpexstr(picturep->ftype);
 if(impex && impex->rendermem)  return(impex->rendermem(picturep,tr,pr,v));
 else                           return(0);
}



int imdatashift(picturestr * picturep)
{
 impexstr   * impex;

 impex=findimpexstr(picturep->ftype);
 if(impex && impex->datashift)  return(impex->datashift(picturep));
 else                           return(0);
}




int impexdata(int ftype)
{
 impexstr * impex;
 impex=findimpexstr(ftype);
 if(impex) return(impex->data);
 else      return(0);
}



int impexdatatrans(int ftype,char * name)
{
 os_error * err;
 impexstr * impex;
	int        i;

 impex=findimpexstr(ftype);
 if(!impex)
 {
  if(name)
  {
 	 for(i=0;i<impexn;i++)
   {
				impex=impexs+i;
    if(impex->data & IMPXPICT)
				{
					ftype=impex->ftype;
   		err=translatetest(name,ftype);
     if(!err) break;
				}
			}

   if(i>=impexn) return(0);
  }
  else
  {
   return(0);    // null name
  }
 }

 return(impex?impex->data:0);
}



int impexftypeindex(int i,int flags)
{
	if(i>=0 && i<impexn)
 {
		if(impexs[i].data & flags) return(impexs[i].ftype);
		return(0);
	}
	return(-1);
}


char * impexname(int ftype)
{
 impexstr * impex;
 impex=findimpexstr(ftype);
 if(impex) return(impex->xtypename);
 else      return(NULL);
}




os_error * dsexport(int ftype,int object,filestr * file,buffer * bf)
{
 os_error * err;
 impexstr * impex;

 err=NULL;

 impex=findimpexstr(ftype);
 if(impex)
 {
  bf_printfeol(bf,"{data \"%s\"",impex->xtypename);

  if(impex->dsexport) err=impex->dsexport(ftype,object,file,bf);

  bf_printfeol(bf,"}");
 }

 return(err);
}



static int       exsaveobject;
static filestr * exsavefile;
static int       extype;

static int       csave;





os_error * export(char * name,int object,int ftype,int xv,filestr * file)
{
 os_error * err;
 impexstr * impex;
 int        type;
 buffer     bf;

 err=NULL;

 type=ftype & IMPFTYPE;
 if(type==DTYPE2)
 {
  impex=findimpexstr(extype);

  if(impex && impex->dsexport) 
  {
   err=bf_open(name,'w',DEFBUFFSIZE,&bf);
   if(!err)
   {
    err=impex->dsexport(ftype,object,file,&bf);
    err=bf_closec(&bf,err,name,DTYPE2);
   }
  }
 }
 else
 {
  impex=findimpexstr(ftype);

  if(impex && impex->export) err=impex->export(name,object,ftype,xv,file);
 }

 return(err);
}


#ifdef NEVER

static os_error * impexsavefn(char * name,int ftype)
{
 return(export(name,exsaveobject,
                    (ftype & IMPFTYPE)|(extype & (~IMPFTYPE)),1,exsavefile));

}



static os_error * impexinternal(mousestr * mouse,char * name,int ftype)
{
 os_error * err;
 impexstr * impex;
 posnstr    posn;

 err=NULL;

 impex=findimpexstr(extype);

 ftype=(ftype & IMPFTYPE)|(extype & (~IMPFTYPE));

 if(impex && impex->intexport) 
 {
  initposn2(&posn,mouse);
  if(posn.valid)
  {
   err=impex->intexport(exsaveobject,ftype,exsavefile,&posn);
  }
 }

 return(err);

 USE(name);
}



static os_error * impexinfo(int * len)
{
 os_error * err;
 impexstr * impex;
 int        type;

 err=NULL;

 type=extype & IMPFTYPE;
 if(type==DTYPE2)
 {
  *len=DEFDSSIZE;
 }
 else
 {
  impex=findimpexstr(type);

  if(impex && impex->exportinfo) 
  {
   err=impex->exportinfo(exsaveobject,extype,exsavefile,len);  
  }
  else 
  {
   *len=0x100;
  }
 }

 return(err);
}

#endif


os_error * impexpxinfo(int object,int ftype,
                                         filestr * file,int * info)
{
 os_error * err;
 impexstr * impex;

 impex=findimpexstr(ftype);

 if(impex && impex->exportinfo) 
 {
  err=impex->pxinfo(object,ftype,file,info);  
 }
 else 
 {
  *info=0;
  err=geterror(ENSFL);;
 }

 return(err);
}










static os_error * saveoptfn(int handle,int opt,int * type,TCHAR * name,int maxlen)
{
 csave=opt;
 if(csave==0)
 {
  tsprintf(256,name,"{TXF0}");
  *type=TEXTFILE;
 }
 else      
 if(csave==1)
 {
  *type=DTYPE2;
  tsprintf(256,name,"{TXF1}");
 }
	else
 if(csave==2)
 {
  *type=UNICODETEXTFILE;
  tsprintf(256,name,"{TXF2}");
 }
 else
 {
  xgetsaveoption(3,opt,name,type);
 }

 return(NULL);
 USE(handle);
}



static saveoptionstr saveoptions[]=
{
 TEXT("{TXF0}"),saveoptfn,
 TEXT("{TXF1}"),saveoptfn,
	TEXT("{TXF2}"),saveoptfn,
};



#ifdef NEVER

static int exporttextramsave(int flags)
{
 if(csave>1) return(1);
 else        return(0);
 USE(flags);
}


#endif


static os_error * impexsavefn2(char * name,int ftype)
{
 os_error * err;
 int        sh;

 if(csave>2)
 {
  err=sc_create(&sh);
  if(!err)
  {
   err=export(sc_name(sh),exsaveobject,
                    (DTYPE2 & IMPFTYPE)|(extype & (~IMPFTYPE)),1,exsavefile);
   if(!err)
   {
    err=translatefile(sc_name(sh),DTYPE2,name,ftype);
    fs_delete(sc_name(sh));
   }
   err=sc_remove(sh,err);
  }
  return(err);
 }
 else
 {
  return(export(name,exsaveobject,
                    (ftype & IMPFTYPE)|(extype & (~IMPFTYPE)),1,exsavefile));
 }
}



#ifdef NEVER

os_error * exportsave(int ftype,int object,filestr * file,int * handle)
{
 os_error * err;
 impexstr * impex;

 err=NULL;

 impex=findimpexstr(ftype);
 if(impex)
 {
  exsaveobject=object;
  exsavefile=file;
  extype=ftype;

  if(impex->data & IMPXTEXT)
  {
   csave=0;
   err=xsetupsaveoptions(saveoptions,sizeof(saveoptions),saveoptfn,&csave);
  }

/*
  if(impex->data & IMPXPICT)
  {
   csave=0;
   err=xsetupsaveoptions(saveoptions,sizeof(saveoptions),saveoptfn,&csave);
  }
*/

  err=setsave(impex->xtypename,ftype & IMPFTYPE,impexsavefn2,handle);
      setsave2(impexinternal,impexinfo);
  if(!err) err=setsave3(exporttextramsave);
 }

 return(err);
}


#endif


os_error * exportsave2(int ftype,int object,viewstr * view,int spare)
{
 os_error * err;
 impexstr * impex;
 TCHAR      name[FILENAMELEN];

 err=NULL;

 impex=findimpexstr(ftype);
 if(impex)
 {
  exsaveobject=object;
  exsavefile=view->file;
  extype=ftype;

  if(impex->data & IMPXTEXT)
  {
   csave=0;
   err=xsetupsaveoptions(saveoptions,sizeof(saveoptions),saveoptfn,&csave);
  }

  xzstrncpyu(name,impex->xtypename,NITEM(name));
  err=opensaveparent(name,NITEM(name),ftype & IMPFTYPE,impexsavefn2,view->handle);
 }

 return(err);

 USE(spare);
}





#ifdef NEVER


static os_error * expdirint(mousestr * mouse,char * name,int ftype)
{
 return(NULL);

 USE(mouse);
 USE(name);
 USE(ftype);
}



os_error * exportdirect(int ftype,int object,filestr * file)
{
 os_error * err;
 impexstr * impex;
 mousestr   mouse;

 err=NULL;

 impex=findimpexstr(ftype);
 if(impex)
 {
  exsaveobject=object;
  exsavefile=file;
  extype=ftype;

  getpointer(&mouse);

  err=setsave0(ftype & IMPFTYPE,impexsavefn);
  setsave2(expdirint,impexinfo);
  err=sendsave2(ftype & IMPFTYPE,impex->xtypename,&mouse);
 }

 return(err);
}

#endif

/****************************************************************************/
// only used for export to clipboard after PASTE_GLOBAL event

os_error * exportglobal(int ftype,int object,filestr * file,char * name)
{
 os_error  * err;
 impexstr  * impex;
 void      * data;
 void      * global;
 int         length;

 err=NULL;

 impex=findimpexstr(ftype);
 if(impex)
 {
  exsaveobject=object;
  exsavefile=file;
  extype=ftype;

  ramnextwritefile(&data,&length);

  err=export(name,exsaveobject,(ftype & IMPFTYPE)|(extype & (~IMPFTYPE)),1,exsavefile);
  if(!err)
  {
			if(data!=NULL && length!=0)
			{
    oscballoc((void**)&global,length);
    memcpy(global,data,length);
			}
  }

//  cprintf("ramwrite off called");
  ramwriteoff();
 }

 return(err);
}



static posnstr * ixposn;
static int       ixtype;
static int       ixflags;
static char    * ixname;


// problem here is that 'length' can't be trusted, it is the
// size of memory allocation

static os_error * importpaste(void * data,int length)
{
 os_error * err;

 cprintf("import paste length %d type %x",length,ixtype);

// fs_saveblock("c:\\Temp\\xx.bmp",BMP,data,length);

 err=NULL;

 if(data && length)
 {
  ramnextreadfile(data,length);

           err=objectload(ixname,ixtype,1,ixposn);
  if(!err) err=objectloadpost(ixposn,ixflags);

  ramreadoff();
 }

 return(err);
}


static os_error * importpastetranslate(void * data,int length)
{
 os_error * err;
	void     * eh;
 int        transhandle;
	char       transname[FILENAMELEN];

 cprintf("import paste translate length %d type %x",length,ixtype);

 err=NULL;

 if(data && length)
 {
  err=sc_create(&transhandle);
  if(!err)
		{
			err=xzstrncpye(transname,sc_name(transhandle),sizeof(transname));
			strcat(transname,".");
			if(!err) err=getextension(ixtype,transname+strlen(transname));

			if(!err) err=fs_saveblock(transname,ixtype,data,length);

//			cprintf("save %s",sc_name(transhandle));
			if(!err)
			{
 		 err=transview(ixflags,transname,ixtype,1,ixposn,TRANS_CONTENT);
			}

//			report(err);

	  eh=errsave(err);
   fs_delete(transname);
   sc_remove(transhandle,err);
   err=errload(eh,NULL);
 	}
 }

 return(err);
}




os_error * importglobal(char * name,posnstr * loadposn,int type,int flags)
{
 os_error * err;
 int        method;
 int        format;

 cprintf("importglobal type %x",type);

	method=0;

 err=objectloadtype(name,type,loadposn,&method);
 if(!err)
 {
		cprintf("loadtype %d",method);

  ixposn=loadposn;
  ixtype=type;
  ixflags=flags;
  ixname=name;

  format=clipgetformat(type);

		cprintf("format %d",format);

  if(method==0)
  { /* assume this means we don't understand the filetype */
    /* get a scrap/real file - then have another go at it */

   // transview(OBCARETLOAD);

  	err=oscbpaste(format,importpastetranslate);
		}
  else
  {
   err=oscbpaste(format,importpaste);
  }
 }

 return(err);
}



/*****************************************************************************/



static os_error * dsdata(int fp,int argc,objectloc * loc)
{              
 os_error        * err;
 filestr         * file;
 dsobjectstr     * ds;
 int               type;

 static int        first;       
 static impexstr * impex;


 err=NULL;

 file=loc->file;

 if(argc==DSIMMINIT)
 {
  first=1;
  impex=NULL;
 }
 else
 if(argc==DSIMMLAST || argc==DSIMMERR)
 {
  if(impex)
  {
   if(impex->dsimport) err=impex->dsimport(impex->ftype,fp,argc,loc);
  }
  else
  {
   ds=(dsobjectstr*)(dsstack+fp);
   ds->data.integer=0;
   ds->len=3*sizeof(int);
  }
 }
 else
 {
  ds=(dsobjectstr*)(dsstack+fp+2);
  type=ds->type;

  if(first)
  {
   if(type==DSTSTR)
   {
    impex=findimpexname(ds->data.string);
    if(impex)
    {
     if(impex->dsimport) err=impex->dsimport(impex->ftype,fp,DSIMMINIT,loc);
    }
   }
   first=0;
  }
  else
  {
   if(impex)
   {
    if(impex->dsimport) err=impex->dsimport(impex->ftype,fp,argc,loc);
   }
  }
 }

 return(err);
}




static int dsblanklist[]={
                     DSTENDLIST
                    };




static dscreator dsdefs[]=
{
 0,0,0,0, "data", dsdata,  DSCREATEIMMEDIATE, DSTDATA,    dsblanklist,
};



/* int impexbits(int ftype,int bic,int orr); */

int impexbitx(int ftype,int bic,int orr)
{
 impexstr * impex;

 impex=findimpexstr(ftype);
 if(impex)
 {
  impex->data&=~bic;
  impex->data|=orr;
  return(impex->data);
 }
 return(-1);
}

int impexbits(int fp)
{
 return(impexbitx(stack[fp],stack[fp+1],stack[fp+2]));
}



/* int impexcopy(int basetype,int newtype,string & name); */

int impexcopy(int fp)
{
 os_error * err;
 impexstr * impex;
 impexstr   impexs;
 char     * p;
 char     * q;

 impex=findimpexstr(stack[fp]);

 if(impex)
 {
  impexs=*impex;
  p=stringptr(stack[fp+2]);
  err=salloc((flex_ptr)&q,strlen(p)+1);
  if(!err)
  {
   strcpy(q,p);
   impexs.xtypename=q;
   impexs.ftype=stack[fp+1];
   impexregister(&impexs);
   return(0);
  }
 }

 return(-1);
}



os_error * impexinit(void)
{

 dsaddblock(dsdefs,sizeof(dsdefs)/sizeof(dscreator));

 return(NULL);
}


