/*->c.dbtext */



#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "akbd.h"
#include "transform.h"
#include "etc.h"
#include "xmath.h"

#include "wos.h"
#include "err.h"
#include "task.h"
#include "xext.h"
#include "poll.h"
#include "key.h"
#include "temp.h"
#include "redraw.h"
#include "trans.h"
#include "fsx.h"
#include "bf.h"


#include "dbhi.h"
#include "mhi.h"



#include "str.h"
#include "object.h"
#include "reslink.h"
#include "units.h"
#include "xcolour.h"
#include "file.h"
#include "posn.h"
#include "select.h"
#include "geometry.h"
#include "pref.h"


#include "view.h"
#include "viewr.h"

#include "txstr.h"
#include "tx.h"
#include "text.h"

#include "dbscale.h"
#include "dbcolour.h"
#include "dbcolumn.h"

#include "dbtext.h"




/*****************************************************************************/

extern dboxstr textbox;

static boxbitstr           ob;
static boxbitstr           nb;
static unpackedtextdatastr pd;

static textchangebits      cbits;


static filestr * cfile;
static viewstr * cview;
static int       cobject;


#define FILLICON    32
#define COLICON     20
#define COLPICKICON 21
#define ALIGNICON   24
#define COLUMNICON  29
#define FIRSTLINEICON 41


static char   *    valignstring[4]={"{VA00}","{VA01}","{VA02}","{VA03}"};
static char   *    firstlinestring[4]={"{FB00}","{FB01}","{FB02}","{FB03}"};


static os_error * colredrawsub(int handle,wimp_redrawstr * redrawstr,int more)
{
 os_error * err;

 err=NULL;

 while(more)
 {
  redrawblock(redrawstr,handle,COLICON,vfindcolour2(NULL,cfile,&pd.colour));
  wimp_get_rectangle(redrawstr,&more);
 }

 return(err);
}


static os_error * textredraw(int handle,int userhandle)
{
 os_error     * err;
 wimp_redrawstr redrawstr;
 int            more;
 redrawstr.w=handle;
 err=wimp_redraw_wind(&redrawstr,&more);
 if(!err) colredrawsub(handle,&redrawstr,more);
 return(err);
 userhandle=userhandle;
}


static void pickclose(colourvalue * colour)
{
 pd.colour=*colour;
 wimp_set_icon_state(textbox.handle,COLICON,
                                    (wimp_iconflags)0,(wimp_iconflags)0);
}


static void columnclose(dbcolumnstr * cdb)
{
 pd.columndata=cdb->ccol;
/* pd.gutter=cdb->gutter; */
}


static os_error * texticon(int handle,int userhandle,wimp_mousestr * m)
{
 os_error     * err;
 dbcolumnstr    cdb;

 err=NULL;

 switch(m->i)
 {
  case COLUMNICON:            /* columns */
          cdb.ccol=pd.columndata;
         /* cdb.gutter=pd.gutter; */
          cdb.width=nb.width;
          cdb.columns=pd.columns;
          cdb.units=cview->data.units;

          err=docolumn(columnclose,&cdb,cfile,&textbox);
          break;

  case FILLICON:            /* fill */
          break;

  case ALIGNICON:
          err=openmenuparent("valign_menu",handle);
          break;

  case FIRSTLINEICON:
          err=openmenuparent("firstline_menu",handle);
          break;

  case COLPICKICON:            /* colour */
          err=pickcolour(cfile,&pd.colour,pickclose,&textbox);
          break;
 }

 return(err);

 USE(handle);
 USE(userhandle);
}


#define MINWIDTH         500
#define MAXWIDTH      2590000
#define MINHEIGHT        500
#define MAXHEIGHT     2590000


static os_error * validtext(boxbitstr * nb,unpackedtextdatastr * pd)
{
 int error;

 error=ENONE;

 if(nb->width<MINWIDTH)    error=BOX01; /* width too small */
 else
 if(nb->width>MAXWIDTH)    error=BOX02; /* width too large */
 else
 if(nb->height<MINHEIGHT)  error=BOX03; /* height too small */
 else
 if(nb->height>MAXHEIGHT)  error=BOX04; /* height too large */
 else
 if(2*pd->inset>nb->width) error=BOX05;  /* Inset too big */
 else
 if(pd->inset<0)           error=BOX06;  /* Inset too small */

 if(error!=ENONE) return(geterror(error));
 else return(columncheck(nb->width,pd->columns,&pd->columndata
                                                       /*,&pd->gutter*/));
}



static os_error * textup(int text,filestr * file)
{
 os_error            * err;
 textboxstr          * textp;

 err=textboxupdate(&pd,&cbits,text,file);

 if(cbits.xtransform) 
 {
  textp=(textboxstr*)fob_find(text,file);
  err=transtext(textp,file,boxtrans,0);
 }

 return(err);
}



static os_error * textclose(int code)
{
 os_error      * err;

 err=NULL;

 if(code==DBOK || code==DBAPPLY) 
 {
  err=validtext(&nb,&pd);
  if(!err)
  {
   textchanges(cobject,cfile,&pd,&cbits);
   if((cbits.xtransform=(cmpboxbits(&ob,&nb)!=0))!=0)
   {
    setboxtrans(&ob,&nb);
    ob=nb;
   }
   modifyselection(textup);
  }
 }
 return(err);
}



static os_error * wralign(int value,char * string)
{
 return(tsprintf(12,string,"%s",valignstring[value & 0x3]));
}

static os_error * wrfirstline(int value,char * string)
{
 return(tsprintf(12,string,"%s",firstlinestring[(value>>2) & 0x3]));
}


static os_error * outu(int value,char * string)
{
 strcpy(string,unitstostring(value,cview->data.units));
 return(NULL);
}


static os_error * inu(int * value,char * string)
{
 return(stringtounits(string,value,cview->data.units));
}


static os_error * outi(int value,char * string)
{
 if(pd.changei) *string=0;
 else           strcpy(string,unitstostring(value,cview->data.units));
 return(NULL);
}


static os_error * ini(int * value,char * string)
{
 if(*string)
 {
  pd.changei=0;
  return(stringtounits(string,value,cview->data.units));
 }
 else
 {
  return(NULL);
 }
}


static os_error * out(int value,char * string)
{
 strcpy(string,numbertostring(value));
 return(NULL);
}


static os_error * in(int * value,char * string)
{
 return(stringtonumber(string,value));
}




static os_error * outux(int value,char * string)
{
 strcpy(string,unitstostring(value-prefp(cfile)->rulerx,cview->data.units));
 return(NULL);
}

static os_error * inux(int * value,char * string)
{
 os_error * err;
 int        n;
 err=stringtounits(string,&n,cview->data.units);
 *value=n+prefp(cfile)->rulerx;
 return(err);
}



static os_error * outuy(int value,char * string)
{
 strcpy(string,unitstostring(value-prefp(cfile)->rulery,cview->data.units));
 return(NULL);
}

static os_error * inuy(int * value,char * string)
{
 os_error * err;
 int        n;
 err=stringtounits(string,&n,cview->data.units);
 *value=n+prefp(cfile)->rulery;
 return(err);
}




static dbiconstr texticondefs[]=
{
 /* N   &V           Type     Grp   Flags   R   L   D   U     In    Out */
   30,  &nb.orgx,    DBWRITE,  0,   0,      -1, -1, 3, -1,    inux, outux,
    3,  &nb.orgy,    DBWRITE,  0,   0,      -1, -1, 5,  1,    inuy, outuy,
    5,  &nb.width,   DBWRITE,  0,   0,      -1, -1, 7,  3,    inu,  outu,
    7,  &nb.height,  DBWRITE,  0,   0,      -1, -1, 9,  5,    inu,  outu,
    9,  &nb.angle,   DBWRITE,  0,   DBASSOCINC,-1, -1,30,  7,readangle,writeangle,
   16,  &nb.skew,    DBWRITE,  0,   DBASSOCINC,-1, -1,30,  7,readskew,writeskew,

   18,  &pd.radius,  DBWRITE,  0,   0,      -1, -1,18,  9,    inu,  outu,
   26,  &pd.inset,   DBWRITE,  0,   0,      -1, -1,24, 30,    ini,  outi,
   34,  &pd.columns, DBWRITE,  0,   DBASSOCINC,-1, -1, 1, 18,    in,   out,

   23,  &pd.valign,  DBTEXT,   1,   0,      0,  0,  0,   0,   NULL,wralign,

   40,  &pd.valign,  DBTEXT,   0,   0,      0,  0,  0,   0,   NULL,wrfirstline,

 /* N   &V           Type     Grp   Flags   Act Key     - -   clickfn 0 */

    1, NULL,         DBACTION, 0,   0,      DBOK,RETURN, 0,0,    NULL, 0,
    2, NULL,         DBACTION, 0,   0,      DBCANCEL,ESCAPE, 0,0,NULL, 0,
   42, NULL,         DBACTION, 0,   0,      DBAPPLY,-1, 0,0,    NULL, 0,

 /* N   &V           Type     Grp   Flags   Step Key  max  min clickfn 0 */

   12, &nb.angle,    DBINC   ,0,    0,       -0x10000,-1,MAXANGLE,
                                                        MINANGLE,NULL,0,

   11, &nb.angle,    DBDEC   ,0,    0,       -0x10000,-1,MAXANGLE,
                                                        MINANGLE,NULL,0,

   37, &nb.skew,     DBINC   ,0,    0,       -0x10000,-1,MAXSKEW,
                                                        MINSKEW,NULL,0,

   36, &nb.skew,     DBDEC   ,0,    0,       -0x10000,-1,MAXSKEW,
                                                        MINSKEW,NULL,0,

   28, &pd.columns,  DBINC   ,0,    0,       1,       -1,MAXCOLUMNS,1,NULL,0,
   27, &pd.columns,  DBDEC   ,0,    0,       1,       -1,MAXCOLUMNS,1,NULL,0,

 /* N   &V           Type     Grp   Flags   Msk Key -  -      clickfn 0 */

   13, &pd.lock,     DBTOGGLE, 0,    0,      1, F4,  0, 0,     NULL,   0,
   14, &pd.spread,   DBTOGGLE, 0,    0,      1, F4,  0, 0,     NULL,   0,

   -1, NULL,         0,        0,    0,      0,  0,  0,  0,  0,      0
};



static dboxstr textbox=
{
 0,
 TXFRAME,
 DBFIX,
 0,
 textclose,
 texticon,
 NULL,
 textredraw,
 texticondefs,
 0,
 0,
 0,NULL,
 0,
 0,
};



/* script function opens the new document dialogue box */

os_error * opentextbox(textboxstr * textp,filestr * file,viewstr * view)
{
 os_error        * err;
 int               regular;

 cfile=file;
 cview=view;
 cobject=textp->key;
 regular=(textp->icurve.type==CBOX);

 unpacktextdata(&pd,textp);

 getboxbits(&textp->fcurve.box,&textp->transform,&ob);
 nb=ob;

 err=dodboxspparent(&textbox,1,&textbox,&dbviewstack,cview->handle);

 dbshade(1,regular?0:DBSHADEG,&textbox);
 shadeiconst(textbox.handle,ALIGNICON,!regular);
 shadeiconst(textbox.handle,22,!regular);

 return(err);
}


/*****************************************************************************/

/* set vertical alignment */

void setvalign(int fp)
{
 pd.valign&=~0x3;
 pd.valign|=(stack[fp] & 0x3);
 dbchangevalue(&pd.valign,&textbox);
}

/* get vertical alignment */

int getvalign(int fp)
{
 return(pd.valign & 0x3);
 USE(fp);
}



void setfirstline(int fp)
{
 pd.valign&=~0xC;
 pd.valign|=((stack[fp] & 0x3)<<2);
 dbchangevalue(&pd.valign,&textbox);
}



int getfirstline(int fp)
{
 return((pd.valign>>2) & 0x3);
 USE(fp);
}



